from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files
from testlib.utils import get_global_asset

import json
from textwrap import dedent


@with_files(
    {
        "multiple-outputs.nix": get_global_asset("multiple-outputs.nix"),
        "config.nix": get_global_asset("config.nix"),
    }
)
def test_regression_6572_independent(nix: Nix):
    """
    https://github.com/NixOS/nix/issues/6572
    """
    nix.settings.add_xp_feature("nix-command")
    independent_text = (
        nix.nix(["build", "-f", "multiple-outputs.nix", "--json", "independent", "--no-link"])
        .run()
        .ok()
    )
    independent_json = json.loads(independent_text.stdout_plain)

    # Make sure that 'nix build' can build a derivation that depends on both outputs of another derivation.
    res = (
        nix.nix(
            [
                "build",
                "-f",
                "multiple-outputs.nix",
                "use-independent",
                "--no-link",
                "--print-out-paths",
            ]
        )
        .run()
        .ok()
    )
    path = res.stdout_plain

    res = nix.nix_store(["--delete", path]).run().ok()
    assert "1 store paths deleted" in res.stderr_plain

    # Make sure that 'nix build' tracks input-outputs correctly when a single output is already present.
    res = nix.nix_store(["--delete", independent_json[0]["outputs"]["first"]]).run().ok()
    assert "1 store paths deleted" in res.stderr_plain

    res = (
        nix.nix(
            [
                "build",
                "-f",
                "multiple-outputs.nix",
                "use-independent",
                "--no-link",
                "--print-out-paths",
            ]
        )
        .run()
        .ok()
    )
    assert nix.physical_store_path_for(res.stdout_plain).read_text() == dedent("""\
        first
        second
    """)

    res = nix.nix_store(["--delete", path]).run().ok()
    assert "1 store paths deleted" in res.stderr_plain

    # Make sure that 'nix build' tracks input-outputs correctly when a single output is already present.
    res = nix.nix_store(["--delete", independent_json[0]["outputs"]["second"]]).run().ok()
    assert "1 store paths deleted" in res.stderr_plain

    res = (
        nix.nix(
            [
                "build",
                "-f",
                "multiple-outputs.nix",
                "use-independent",
                "--no-link",
                "--print-out-paths",
            ]
        )
        .run()
        .ok()
    )
    assert nix.physical_store_path_for(res.stdout_plain).read_text() == dedent("""\
        first
        second
    """)


@with_files(
    {
        "multiple-outputs.nix": get_global_asset("multiple-outputs.nix"),
        "config.nix": get_global_asset("config.nix"),
    }
)
def test_regression_6572_dependent(nix: Nix):
    """
    https://github.com/NixOS/nix/issues/6572
    """
    nix.settings.add_xp_feature("nix-command")
    res = nix.nix(["build", "-f", "multiple-outputs.nix", "--json", "a", "--no-link"]).run().ok()
    second_path = json.loads(res.stdout_plain)[0]["outputs"]["second"]

    path = (
        nix.nix(["build", "-f", "multiple-outputs.nix", "use-a", "--no-link", "--print-out-paths"])
        .run()
        .ok()
        .stdout_plain
    )
    res = nix.nix_store(["--delete", path]).run().ok()
    assert "1 store paths deleted" in res.stderr_plain

    res = nix.nix_store(["--delete", second_path]).run().ok()
    assert "1 store paths deleted" in res.stderr_plain

    res = (
        nix.nix(["build", "-f", "multiple-outputs.nix", "use-a", "--no-link", "--print-out-paths"])
        .run()
        .ok()
    )
    assert nix.physical_store_path_for(res.stdout_plain).read_text() == dedent("""\
        first
        second
    """)



import pytest
import sys

from testlib.fixtures.file_helper import with_files, File
from testlib.fixtures.nix import Nix
from testlib.xattrs import verify_no_xattrs_in_tree, skip_if_xattrs_are_unsupported

import xattr
import json
import re


@with_files({"file-with-xattrs": File("hi")})
@pytest.mark.parametrize("hash_fn", ["sha1", "sha256"])
@pytest.mark.skipif(
    sys.platform != "linux", reason="xattrs scrubbing is not expected to function outside of Linux"
)
def test_add_xattrs_fixed_flat_file(nix: Nix, hash_fn: str):
    skip_if_xattrs_are_unsupported(nix.env)

    xattr.set(nix.env.dirs.home / "file-with-xattrs", "user.deriver", "test")

    res = nix.nix_store(["--add-fixed", hash_fn, "file-with-xattrs"]).run().ok()
    verify_no_xattrs_in_tree(nix.physical_store_path_for(res.stdout_plain))


@with_files({"file-with-xattrs": File("hi")})
@pytest.mark.skipif(
    sys.platform != "linux", reason="xattrs scrubbing is not expected to function outside of Linux"
)
def test_add_xattrs_flat_file(nix: Nix):
    skip_if_xattrs_are_unsupported(nix.env)

    xattr.set(nix.env.dirs.home / "file-with-xattrs", "user.deriver", "test")

    res = nix.nix_store(["--add", "file-with-xattrs"]).run().ok()
    verify_no_xattrs_in_tree(nix.physical_store_path_for(res.stdout_plain))


@with_files({"dir-with-xattrs": {"test1": File("hi")}})
@pytest.mark.parametrize("hash_fn", ["sha1", "sha256"])
@pytest.mark.skipif(
    sys.platform != "linux", reason="xattrs scrubbing is not expected to function outside of Linux"
)
def test_add_xattrs_directories(nix: Nix, hash_fn: str):
    skip_if_xattrs_are_unsupported(nix.env)

    xattr.set(nix.env.dirs.home / "dir-with-xattrs", "user.deriver", "hi")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs/test1", "user.deriver", "hi")

    res = nix.nix_store(["--add-fixed", hash_fn, "--recursive", "dir-with-xattrs"]).run().ok()
    verify_no_xattrs_in_tree(nix.physical_store_path_for(res.stdout_plain))


@with_files({"file-with-xattrs": File("hi"), "dir-with-xattrs": {"test1": File("hi")}})
@pytest.mark.skipif(
    sys.platform != "linux", reason="xattrs scrubbing is not expected to function outside of Linux"
)
def test_interpolating_xattrs_file(nix: Nix):
    skip_if_xattrs_are_unsupported(nix.env)

    xattr.set(nix.env.dirs.home / "file-with-xattrs", "user.deriver", "test")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs", "user.deriver", "hi")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs/test1", "user.deriver", "hi")

    res = nix.eval('"${./file-with-xattrs}"', flags=["--impure"]).ok()
    store_path = json.loads(res.stdout_plain)
    verify_no_xattrs_in_tree(nix.physical_store_path_for(store_path))

    res = nix.eval('"${./dir-with-xattrs}"', flags=["--impure"]).ok()
    store_path = json.loads(res.stdout_plain)
    verify_no_xattrs_in_tree(nix.physical_store_path_for(store_path))


@with_files({"file-with-xattrs": File("hi"), "dir-with-xattrs": {"test1": File("hi")}})
@pytest.mark.skipif(
    sys.platform != "linux", reason="xattrs scrubbing is not expected to function outside of Linux"
)
def test_builtins_path_xattrs(nix: Nix):
    skip_if_xattrs_are_unsupported(nix.env)

    xattr.set(nix.env.dirs.home / "file-with-xattrs", "user.deriver", "test")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs", "user.deriver", "hi")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs/test1", "user.deriver", "hi")

    res = nix.eval(
        'builtins.path { name = "test-file"; path = ./file-with-xattrs; }', flags=["--impure"]
    ).ok()
    store_path = json.loads(res.stdout_plain)
    verify_no_xattrs_in_tree(nix.physical_store_path_for(store_path))

    res = nix.eval(
        'builtins.path { name = "test-file"; path = ./file-with-xattrs; recursive = true; }',
        flags=["--impure"],
    ).ok()
    store_path = json.loads(res.stdout_plain)
    verify_no_xattrs_in_tree(nix.physical_store_path_for(store_path))

    res = nix.eval(
        'builtins.path { name = "test-dir"; path = ./dir-with-xattrs; recursive = true; }',
        flags=["--impure"],
    ).ok()
    store_path = json.loads(res.stdout_plain)
    verify_no_xattrs_in_tree(nix.physical_store_path_for(store_path))


@with_files({"file-with-xattrs": File("hi"), "dir-with-xattrs": {"test1": File("hi")}})
@pytest.mark.skipif(
    sys.platform != "linux", reason="xattrs scrubbing is not expected to function outside of Linux"
)
def test_nix_prefetch_xattrs(nix: Nix):
    skip_if_xattrs_are_unsupported(nix.env)

    xattr.set(nix.env.dirs.home / "file-with-xattrs", "user.deriver", "test")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs", "user.deriver", "hi")
    xattr.set(nix.env.dirs.home / "dir-with-xattrs/test1", "user.deriver", "hi")

    locate_path_re = re.compile(r"path is '(.*)'")
    for p in ("file-with-xattrs", "dir-with-xattrs"):
        res = nix.nix_prefetch_url([f"file://{nix.env.dirs.home / p}"]).run().ok()
        matches = locate_path_re.findall(res.stderr_plain)
        assert matches, "nix prefetch must return at least one path"
        verify_no_xattrs_in_tree(nix.physical_store_path_for(matches[0]))



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File
from testlib.environ import environ
from testlib.utils import get_global_asset_pack, get_global_asset
from pathlib import Path
import pytest

system = environ.get("system")

files = {
    "templates": get_global_asset_pack(".git")
    | {
        "flake.nix": File("""{
          description = "Some templates";

          outputs = { self }: {
            templates = rec {
              trivial = {
                path = ./trivial;
                description = "A trivial flake";
                welcomeText = ''
                    Welcome to my trivial flake
                '';
              };
              default = trivial;
            };
          };
        }"""),
        "trivial": {
            "flake.nix": File(f"""{{
              description = "A flake for building Hello World";

              outputs = {{ self, nixpkgs }}: {{
                packages.{system} = rec {{
                  hello = nixpkgs.legacyPackages.{system}.hello;
                  default = hello;
                }};
              }};
            }}"""),
            "a": File("a"),
            "b": File("b"),
        },
    },
    "flake": get_global_asset_pack(".git"),
    "nixpkgs": get_global_asset_pack(".git")
    | {
        "config.nix": get_global_asset("config.nix"),
        "flake.nix": File(f"""{{
          outputs = {{ self }}: {{
            legacyPackages.{system}.hello =
              with import ./config.nix;
              mkDerivation {{
                name = "hello";
                buildCommand = "echo hello > $out";
              }};
          }};
        }}"""),
    },
    "registry.json": File(""),
}


@pytest.fixture(autouse=True)
def common_init(nix: Nix, files: Path, git: Git):
    nix.settings.add_xp_feature("nix-command", "flakes")
    nix.settings["flake-registry"] = str(files / "registry.json")

    git(files / "templates", "add", "flake.nix", "trivial/")
    git(files / "templates", "commit", "-m", "Initial")
    git(files / "nixpkgs", "add", "flake.nix", "config.nix")

    nix.nix(
        [
            "registry",
            "add",
            "--registry",
            files / "registry.json",
            "templates",
            f"git+file://{files}/templates",
        ]
    ).run().ok()
    nix.nix(
        [
            "registry",
            "add",
            "--registry",
            files / "registry.json",
            "nixpkgs",
            f"git+file://{files}/nixpkgs",
        ]
    ).run().ok()


@with_files(files)
class TestFlakeTemplates:
    def test_check(self, nix: Nix, files: Path):
        assert (
            "checking template 'templates.trivial'"
            in nix.nix(["flake", "check", files / "templates"]).run().ok().stderr_s
        )

    def test_show(self, nix: Nix, files: Path):
        assert (
            "A trivial flake" in nix.nix(["flake", "show", files / "templates"]).run().ok().stdout_s
        )

    def test_show_json(self, nix: Nix, files: Path):
        result = nix.nix(["flake", "show", files / "templates", "--json"]).run().json()
        assert result["templates"]["default"]["description"] == "A trivial flake"

    def test_init(self, nix: Nix, files: Path):
        stderr = nix.nix(["flake", "init"], cwd=files / "flake").run().ok().stderr_s
        assert f"wrote: {files}/flake/flake.nix" in stderr

    class TestWithInit:
        @pytest.fixture(autouse=True)
        def init_flake(self, nix: Nix, files: Path):
            nix.nix(["flake", "init"], cwd=files / "flake").run().ok()

        def test_init_idempotent(self, nix: Nix, files: Path):
            stderr = nix.nix(["flake", "init"], cwd=files / "flake").run().ok().stderr_s
            assert "skipping identical file" in stderr

        class TestInRepo:
            @pytest.fixture(autouse=True)
            def init_repo(self, files: Path, git: Git):
                git(files / "flake", "add", "flake.nix")

            def test_check(self, nix: Nix, files: Path):
                stderr = nix.nix(["flake", "check", files / "flake"]).run().ok().stderr_s
                assert f"fetching git input 'git+file://{files}/flake'" in stderr
                assert f"fetching git input 'git+file://{files}/nixpkgs'" in stderr
                assert "evaluating flake" in stderr

            def test_show(self, nix: Nix, files: Path):
                stdout = nix.nix(["flake", "show", files / "flake"]).run().ok().stdout_s
                assert "hello: package 'hello'" in stdout

            def test_show_json(self, nix: Nix, files: Path):
                result = nix.nix(["flake", "show", files / "flake", "--json"]).run().json()
                assert result["packages"][system]["default"]["name"] == "hello"
                assert result["packages"][system]["hello"]["name"] == "hello"

    def test_benign_conflict_idempotence(self, nix: Nix, files: Path):
        (files / "flake/a").write_text("a")

        stderr = nix.nix(["flake", "init"], cwd=files / "flake").run().ok().stderr_s
        assert "skipping identical file" in stderr
        assert f"wrote: {files}/flake/flake.nix" in stderr

    def test_init_conflict(self, nix: Nix, files: Path):
        (files / "flake/a").write_text("b")

        stderr = nix.nix(["flake", "init"], cwd=files / "flake").run().expect(1).stderr_s
        assert f"refusing to overwrite existing file '{files}/flake/a'" in stderr

    def test_new(self, nix: Nix):
        stderr = nix.nix(["flake", "new", "-t", "templates#trivial", "flake2"]).run().ok().stderr_s
        assert f"wrote: {nix.env.dirs.home}/flake2/flake.nix" in stderr

    class TestNew:
        def test_idempotent(self, nix: Nix, files: Path):
            nix.nix(["flake", "new", "-t", "templates#trivial", files / "flake2"]).run().ok()
            stderr = (
                nix.nix(["flake", "new", "-t", "templates#trivial", files / "flake2"])
                .run()
                .ok()
                .stderr_s
            )
            assert "skipping identical file" in stderr

        def test_check(self, nix: Nix, files: Path):
            nix.nix(["flake", "new", "-t", "templates#trivial", files / "flake2"]).run().ok()
            stderr = nix.nix(["flake", "check", files / "flake2"]).run().ok().stderr_s
            assert "evaluating flake" in stderr



"""
HTTP server fixture for tests which binds to an auto-assigned port on localhost.
"""

import asyncio
import contextlib
import dataclasses
import logging
import time
import socket
import threading
from queue import Queue
import aiohttp.web as web
import aiohttp.http_writer as http_writer

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class HttpServer:
    app: web.Application
    port: int


class EventTS(asyncio.Event):
    """
    A thread safe version of the asyncio Event

    NOTE: clear() is not thread safe

    Taken from https://stackoverflow.com/a/33006667
    """

    def __init__(self, *args, loop: asyncio.AbstractEventLoop | None = None, **kwargs):
        """
        Creates a thread-safe event for the given loop (or the loop of the current thread).
        """
        super().__init__(*args, **kwargs)
        self.target_loop = loop or asyncio.get_running_loop()

    def set(self):
        self.target_loop.call_soon_threadsafe(super().set)


def _make_localhost_socket(port: int = 0) -> tuple[socket.socket, int]:
    """Creates a localhost-bound socket with an auto-assigned port (unless specified)."""
    sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    # Shouldn't matter because we dynamically allocate ports, but this is generally preferred.
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(("::1", port))
    _, port = sock.getsockname()[:2]

    return (sock, port)


def _server_thread(app: web.Application, sock: socket.socket, shutdown_ev_q: Queue):
    async def async_main():
        nonlocal app, sock
        # Due to Reasons(tm) of event loop lifecycles and stuff of the sort,
        # it's far easier to just send the event object to the other thread
        # from inside the loop where it already knows which loop it is.
        shutdown_ev = EventTS()
        shutdown_ev_q.put(shutdown_ev)

        runner = web.AppRunner(app, handle_signals=False)
        await runner.setup()
        site = web.SockSite(runner, sock)
        await site.start()
        await shutdown_ev.wait()
        await runner.cleanup()

    asyncio.run(async_main())


@contextlib.contextmanager
def http_server(app: web.Application, port: int = 0):
    """
    Creates an http server on an automatically chosen port (if not given) on
    the host running the given web.Application, gives you the port for it.

    The server is run on a separate thread.
    """
    # n.b. pytest doesn't directly support asyncio. There's a bunch of
    # complexity that we could go through to do this or we could just throw the
    # async on a thread which was what we would do to the web server anyway if
    # it was blocking.
    shutdown_ev_q = Queue()
    thr = None
    sock = None
    shutdown_ev = None
    try:
        sock, port = _make_localhost_socket(port=port)
        thr = threading.Thread(
            target=_server_thread,
            args=(app, sock, shutdown_ev_q),
            name=f"functional2 httpd [::1]:{port}",
        )
        thr.start()
        shutdown_ev = shutdown_ev_q.get()
        yield HttpServer(app=app, port=port)
    finally:
        if shutdown_ev:
            shutdown_ev.set()
        if thr:
            thr.join()
        if sock:
            sock.close()


def dev_main():
    """A little test server for poking at this manually"""
    logging.basicConfig(level=logging.DEBUG)

    async def root(_req: web.Request) -> web.Response:
        # sadly required to make this function async
        await asyncio.sleep(0.01)
        return web.Response(body="hello world")

    async def eof(req: web.Request) -> web.StreamResponse:
        # sadly required to make this function async
        await asyncio.sleep(0.01)
        writer: http_writer.StreamWriter = req.writer  # type: ignore
        assert writer.transport
        writer.transport.close()
        return web.StreamResponse()

    async def eof_after_headers(req: web.Request) -> web.StreamResponse:
        resp = web.StreamResponse()
        resp.headers["content-length"] = "1337"
        await resp.prepare(req)
        await resp.write_eof()
        writer: http_writer.StreamWriter = req.writer  # type: ignore
        assert writer.transport
        writer.transport.close()
        return resp

    app = web.Application()
    app.add_routes(
        [web.get("/", root), web.get("/eof", eof), web.get("/eof_after_headers", eof_after_headers)]
    )

    with http_server(app, 1337) as httpd:
        logger.info("Listening on http://[::1]:%d", httpd.port)
        while True:
            time.sleep(3600)


if __name__ == "__main__":
    dev_main()



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File
from testlib.utils import get_global_asset_pack
from testlib.environ import environ
from pathlib import Path
from .common import simple_flake
import pytest

system = environ.get("system")
files = simple_flake() | {
    "b-low": simple_flake()
    | {
        "flake.nix": File(f"""{{
          outputs = inputs: rec {{
            packages.{system} = rec {{
              default =
                assert builtins.readFile ./message == "all good\n";
                assert builtins.readFile (inputs.self + "/message") == "all good\n";
                import ./simple.nix;
            }};
          }};
        }}"""),
        "message": File("all good\n"),
    }
}


@pytest.fixture(autouse=True)
def common_init(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")


@with_files(files)
def test_subdir_self_path(nix: Nix, files: Path):
    assert "simple> PATH=" in nix.nix(["build", f"{files}/b-low", "--no-link"]).run().ok().stderr_s


@with_files({"repo": files | get_global_asset_pack(".git"), "client": {}})
def test_git_subdir_self_path(nix: Nix, files: Path, git: Git):
    repo_dir = files / "repo"
    git(repo_dir, "add", ".")
    git(repo_dir, "commit", "-m", "init")

    client_dir = files / "client"
    (client_dir / "flake.nix").write_text(f"""{{
      inputs.inp = {{
        type = "git";
        url = "file://{repo_dir}";
        dir = "b-low";
      }};

      outputs = inputs: rec {{
        packages = inputs.inp.packages;
      }};
    }}""")

    assert "simple> PATH=" in nix.nix(["build", client_dir, "--no-link"]).run().ok().stderr_s



import pytest
from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import File, with_files


@with_files(
    {
        "flake.nix": File("""{
            outputs = a: {
               packages.system = {
                 foo = 1;
                 fo1 = 1;
                 fo2 = 1;
                 fooo = 1;
                 foooo = 1;
                 fooooo = 1;
                 fooooo1 = 1;
                 fooooo2 = 1;
                 fooooo3 = 1;
                 fooooo4 = 1;
                 fooooo5 = 1;
                 fooooo6 = 1;
               };
            };
        }""")
    }
)
class TestSuggestions:
    @pytest.fixture(autouse=True)
    def setup(self, nix: Nix):
        nix.settings.add_xp_feature("nix-command", "flakes")
        nix.settings.system = "system"

    @pytest.mark.parametrize(
        "args",
        [
            [".#fob"],
            ["--impure", "--expr", "(builtins.getFlake (builtins.toPath ./.)).packages.system.fob"],
        ],
    )
    def test_three_closest(self, nix: Nix, args: list[str]):
        err = nix.nix(["build", *args]).run().expect(1).stderr_s
        assert "Did you mean one of fo1, fo2, foo or fooo?" in err

    def test_only_suggest_relevant(self, nix: Nix):
        assert "Did you mean" not in nix.nix(["build", ".#bar"]).run().expect(1).stderr_s

    def test_inactive_if_all_args_provided(self, nix: Nix):
        assert "Did you mean" not in (
            nix.nix(["build", "--impure", "--expr", "({ foo }: foo) { foo = 1; fob = 2; }"])
            .run()
            .expect(1)
            .stderr_s
        )

    def test_expr(self, nix: Nix):
        assert "Did you mean foo?" in (
            nix.nix(["build", "--impure", "--expr", "({ foo ? 1 }: foo) { fob = 2; }"])
            .run()
            .expect(1)
            .stderr_s
        )



from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files, File
from testlib.fixtures.command import CommandResult
from pathlib import Path
from .common import simple_flake
import tarfile
import pytest

flake_nix = """
{
  inputs.flake2.url = "file://$TEST_ROOT/flake.tar.gz";

  outputs = { self, flake2 }: {

    a1 = builtins.fetchTarball {
      #type = "tarball";
      url = "file://$TEST_ROOT/flake.tar.gz";
      sha256 = "$hash";
    };

    a2 = ./foo;

    a3 = ./.;

    a4 = self.outPath;

    # FIXME
    a5 = self;

    a6 = flake2.outPath;

    # FIXME
    a7 = "${flake2}/config.nix";

    # This is only allowed in impure mode.
    a8 = builtins.storePath $dep;

    a9 = "$dep";

    drvCall = with import ./config.nix; mkDerivation {
      name = "simple";
      builder = ./simple.builder.sh;
      PATH = "";
      goodPath = path;
    };

    a10 = builtins.unsafeDiscardOutputDependency self.drvCall.drvPath;

    a11 = self.drvCall.drvPath;

    a12 = self.drvCall.outPath;

    a13 = "${self.drvCall.drvPath}${self.drvCall.outPath}";
  };
}
"""

files = {
    "flake1": simple_flake() | {"foo": File("bar")},
    "flake2": simple_flake(),
    "common.sh": File("nothing important"),
}


@with_files(files)
class TestBuild:
    @pytest.fixture(autouse=True)
    def common_init(self, nix: Nix, files: Path):
        self.nix = nix
        self.out = nix.env.dirs.home / "result"

        nix.settings.add_xp_feature("nix-command", "flakes")

        with tarfile.open(files / "flake.tar.gz", "w:gz") as tar:

            def drop_prefix(ti: tarfile.TarInfo) -> tarfile.TarInfo:
                ti.name = ti.name.replace(str(files)[1:] + "/", "")
                return ti

            tar.add(files / "flake2", filter=drop_prefix)

        flake_hash = nix.hash_path(files / "flake2")
        dep = nix.nix(["store", "add-path", files / "common.sh"]).run().ok().stdout_s.strip()

        flake = files / "flake1/flake.nix"
        flake.write_text(
            flake_nix.replace("$TEST_ROOT", str(files))
            .replace("$hash", flake_hash)
            .replace("$dep", dep)
        )

    def build(self, *args) -> CommandResult:
        return self.nix.nix(["build", "--json", "--out-link", self.out, *args]).run()

    @pytest.mark.parametrize("attr", ["a1", "a6"])
    def test_build_simple_nix(self, files: Path, attr: str):
        self.build(f"{files}/flake1#{attr}").ok()
        assert (self.out / "simple.nix").exists()

    def test_build_a2(self, files: Path):
        self.build(f"{files}/flake1#a2").ok()
        assert self.out.read_text() == "bar"

    @pytest.mark.parametrize("attr", ["a3", "a4"])
    def test_build_plain(self, files: Path, attr: str):
        self.build(f"{files}/flake1#{attr}").ok()

    def test_build_a8(self, files: Path):
        self.build("--impure", f"{files}/flake1#a8").ok()
        assert self.out.read_text() == "nothing important"

    def test_build_a9(self, files: Path):
        result = self.build(f"{files}/flake1#a9").expect(1)
        assert (
            "has 0 entries in its context. It should only have exactly one entry" in result.stderr_s
        )

    def test_build_a10(self, files: Path):
        self.build(f"{files}/flake1#a10").ok()
        assert str(self.out.readlink()).endswith("simple.drv")

    def test_build_a11(self, files: Path):
        result = self.build(f"{files}/flake1#a11").expect(1)
        assert (
            "has a context which refers to a complete source and binary closure" in result.stderr_s
        )

    def test_build_a12(self, files: Path):
        self.build(f"{files}/flake1#a12").ok()
        assert (self.out / "hello").exists()

    def test_build_a13(self, files: Path):
        result = self.build("--impure", f"{files}/flake1#a13").expect(1)
        assert (
            "has 2 entries in its context. It should only have exactly one entry" in result.stderr_s
        )



from testlib.fixtures.nix import Nix


def test_err_context(nix: Nix):
    # the lang test framework doesn't check this folder, as there is a custom test in here
    # it won't scream about missing an `in.nix` or .exp files
    result = nix.nix_instantiate(
        ["--show-trace", "--eval", "-E", 'builtins.addErrorContext "Hello" (throw "Foo")']
    ).run()
    assert "Hello" in result.expect(1).stderr_s



import contextlib
import copy
import dataclasses
import sys
from functools import partialmethod
from pathlib import Path
from typing import Any
from collections.abc import Callable, Generator
import shutil
import subprocess
import logging

import pytest

from testlib.fixtures.command import CommandResult, Command
from testlib.fixtures.env import ManagedEnv
from testlib.utils import is_value_of_type


type _NixSettingValue = str | int | list[str] | bool | None


def _serialise(value: Any) -> str:
    if is_value_of_type(value, list[str]):
        return " ".join(_serialise(e) for e in value)
    if is_value_of_type(value, bool):
        return "true" if value else "false"
    if is_value_of_type(value, str | int):
        return str(value)

    msg = f"Value is unsupported in nix config: {value!r}, must be {_NixSettingValue.__value__}"
    raise ValueError(msg)


class NixSettings:
    """Settings for invoking Nix"""

    def __init__(self):
        self._settings: dict[str, _NixSettingValue] = {
            # Running the test suite creates a lot of stores in the test root (somewhere under TMPDIR).
            # Obviously, they are not critical for system operation, so there is no need to reserve space.
            # The cleanup will only happen a couple of runs later, wasting space in the meantime.
            # Effectively disable this space reserve to reduce the waste considerably (by about 98%).
            "gc-reserved-space": 0,
            "show-trace": True,
            "sandbox": True,
            # explicitly disable substitution by default, otherwise we may attempt to contact
            # substituters and slow down many tests with pointless connection retry timeouts.
            "substituters": [],
            "extra-sandbox-paths": [],
            "extra-experimental-features": [],
            "extra-deprecated-features": [],
        }

    def __getattr__(self, attr: str) -> _NixSettingValue:
        if attr.startswith("__"):
            return super().__getattr__(attr)
        return self._settings[attr.replace("_", "-")]

    def __setattr__(self, attr: str, value: _NixSettingValue):
        if attr == "_settings":
            super().__setattr__(attr, value)
        else:
            self._settings[attr.replace("_", "-")] = value

    def __getitem__(self, attr: str) -> _NixSettingValue:
        return self._settings[attr]

    def __setitem__(self, attr: str, value: str):
        self._settings[attr] = value

    def add_xp_feature(self, *names: str):
        self["extra-experimental-features"] += names

    def add_dp_feature(self, *names: str):
        self["extra-deprecated-features"] += names

    def update(self, args: dict[str, _NixSettingValue] | None = None, **kwargs):
        """
        Overrides the settings with the given dict or kwargs.
        """
        self._settings.update(kwargs | (args or {}))

    def clone(self) -> "NixSettings":
        """
        shortcut to clone the settings to a new object
        """
        return self.with_settings()

    def with_settings(
        self, args: dict[str, _NixSettingValue] | None = None, **kwargs
    ) -> "NixSettings":
        """
        Copies the current settings into a new object, overriding the provided ones.

        :returns: A new Settings object with overridden settings
        """
        new_settings = NixSettings()
        new_settings._settings = copy.deepcopy(self._settings)
        new_settings.update(args, **kwargs)
        return new_settings

    def to_config(self, env: ManagedEnv) -> str:
        config = ""

        self["extra-sandbox-paths"] += env.path.to_sandbox_paths()

        def field_may(name: str, value: Any, serializer: Callable[[Any], str] = _serialise):
            nonlocal config
            if value is not None:
                config += f"{name} = {serializer(value)}\n"

        for name, value in self._settings.items():
            field_may(name, value)

        return config

    def to_env_overlay(self, env: ManagedEnv) -> None:
        cfg = self.to_config(env)
        (env.dirs.nix_conf_dir / "nix.conf").write_text(cfg)
        env.set_env("NIX_CONFIG", cfg)


@dataclasses.dataclass
class Nix:
    env: ManagedEnv
    logger: logging.Logger
    _settings: NixSettings | None = dataclasses.field(init=False, default=None)

    @property
    def _nix_executable(self) -> Path:
        if nix_bin_dir := self.env.dirs.nix_bin_dir:
            return Path(nix_bin_dir) / "nix"

        if from_path := shutil.which("nix"):
            return Path(from_path)

        raise ValueError(
            "Couldn't find a Nix command to execute! Set NIX_BIN_DIR or fix your environment"
        )

    @property
    def settings(self) -> NixSettings:
        """
        :return: the settings for the nix instance
        """
        if self._settings is None:
            self._settings = NixSettings()
            self._settings.store = f"local?root={self.env.dirs.test_root}"
            if sys.platform == "linux":
                # sandbox build dir cannot be withing store dir. choose a short non-overlapping path.
                self._settings.sandbox_build_dir = (
                    "/build-f2" if self.env.dirs.test_root.parts[1] != "build-f2" else "/build.f2"
                )

        return self._settings

    def nix_cmd(self, argv: list[str], flake: bool = False, cwd: Path | None = None) -> Command:
        """
        Constructs a NixCommand with the appropriate settings.
        :param build: if the executed command wants to build stuff. This is required due to darwin shenanigans. "auto" will try to autodetect, override using `True` or `False`. Has no effect on linux.
        """
        # Create a copy of settings to not have a writing side effect
        settings = self.settings.clone()
        if flake:
            settings.add_xp_feature("nix-command", "flakes")

        settings.to_env_overlay(self.env)
        return Command(argv=argv, exe=self._nix_executable, _env=self.env, cwd=cwd)

    def nix(
        self, cmd: list[str], nix_exe: str = "nix", flake: bool = False, cwd: Path | None = None
    ) -> Command:
        return self.nix_cmd([nix_exe, *cmd], flake=flake, cwd=cwd)

    @contextlib.contextmanager
    def daemon(
        self,
        args: list[str] | None = None,
        settings: dict[str, _NixSettingValue] | None = None,
        **kwargs,
    ) -> "Nix":
        daemon = copy.deepcopy(self)
        daemon.logger = self.logger.getChild("daemon")
        daemon.settings["allowed-users"] = ["*"]
        daemon.settings["trusted-users"] = []
        daemon.settings.store = f"local?root={self.env.dirs.test_root}"
        daemon.settings.update(settings)

        sockets_dir = Path(daemon.env.dirs.nix_state_dir) / "daemon-socket"
        sockets = [sockets_dir / "socket"]
        for p in sockets:
            p.unlink(missing_ok=True)

        proc = daemon.nix(args or [], nix_exe="nix-daemon", **kwargs).start()

        def log_daemon_result(result: CommandResult | None, level: int):
            if result:
                daemon.logger.log(level, "daemon exited with code %i", result.rc)
                daemon.logger.log(level, "stdout: %s", result.stdout_s)
                daemon.logger.log(level, "stderr: %s", result.stderr_s)
            else:
                daemon.logger.error("daemon exited unexpectedly")

        # wait for daemon to come up. this may take a while under load.
        # we only test the *last* socket in the list because that's the
        # last one the daemon creates, once it's there the daemon is up
        while not sockets[-1].exists():
            if status := proc.wait(0.01):
                log_daemon_result(status, logging.ERROR)
                raise RuntimeError("daemon exited during startup")

        inner = copy.deepcopy(self)
        inner.settings.store = f"unix://{sockets[-1]}"  # missing multi socket support

        try:
            timeout, level = 1, logging.ERROR
            yield inner
            # 5 seconds should be enough to wait for a *graceful* exit.
            timeout, level = 5, logging.DEBUG
        finally:
            result = proc.terminate(timeout)
            if not result:
                result = proc.kill()
            log_daemon_result(result, level)

    # Mark each of these as correct as they are not ClassVars, but we also don't want to turn off RUF045
    nix_build = partialmethod(nix, nix_exe="nix-build")  # noqa: RUF045
    nix_shell = partialmethod(nix, nix_exe="nix-shell")  # noqa: RUF045
    nix_store = partialmethod(nix, nix_exe="nix-store")  # noqa: RUF045
    nix_env = partialmethod(nix, nix_exe="nix-env")  # noqa: RUF045
    nix_instantiate = partialmethod(nix, nix_exe="nix-instantiate")  # noqa: RUF045
    nix_channel = partialmethod(nix, nix_exe="nix-channel")  # noqa: RUF045
    nix_prefetch_url = partialmethod(nix, nix_exe="nix-prefetch-url")  # noqa: RUF045

    def eval(
        self, expr: str, settings: NixSettings | None = None, flags: list[str] | None = None
    ) -> CommandResult:
        """
        calls `nix eval --json --expr {expr}` using the given expression
        :param expr: what to evaluate
        :param settings: if none, the global settings will be used, otherwise the given one
        :param flags: if none, empty list, otherwise pass flags to the CLI invocation
        :return: result of the evaluation
        """
        if flags is None:
            flags = []
        orig = self.settings.clone()
        self._settings = settings or self.settings
        self.settings.add_xp_feature("nix-command")

        cmd = self.nix(["eval", "--json", *flags, "--expr", expr])
        # restore previous settings
        self._settings = orig
        return cmd.run()

    @property
    def store_dir(self) -> Path:
        """
        The actual NIX_STORE_DIR this Nix command uses.
        """
        assert self.env.dirs.real_store_dir is not None, "bug in ManagedEnv"
        return self.env.dirs.real_store_dir

    def physical_store_path_for(self, path: str | Path) -> Path:
        """
        Takes a /nix/store/… path and rewrites it to be relative to this Nix's NIX_STORE_DIR.

        Nix accepts and returns store paths as `/nix/store` even when that's not where `NIX_STORE_DIR`
        physically is on the filesystem. Since we move the conceptual root for Nix to `test_root`,
        these "virtual" paths differ from the physical ones. So this function will convert `/nix/store`
        "virtual" paths to their real, physical location on the system.

        Basically, if you're passing it to `nix build` or `nix-store` or whatever, you want the
        `/nix/store` version. If you're passing it to a Python API (like pathlib.Path.exists()) or a
        command that operates on arbitrary files instead of store paths, you want the output of this
        function.


        :param path: a string or Path to convert
        :return: a Path object holding the rewritten, physical system path to the store entry
        """
        return (
            Path(str(path).replace("/nix/store", self.store_dir.as_posix()))
            if str(path).startswith("/nix/store")
            else Path(path)
        )

    def hash_path(self, store_path: str | Path, *args: str) -> str:
        """
        Shortcut to use `nix hash path {store_path}`, converting "virtual" store paths returned
        from Nix to their physical system paths including the test root.

        :param store_path: store path of the derivation or entry to hash
        """
        actual_path = self.physical_store_path_for(store_path).as_posix()
        res = self.nix(["hash", "path", actual_path, *args], flake=True).run().ok()
        return res.stdout_plain

    def clear_store(self):
        """
        Clears the test-owned store (and state) and resets them to an empty state
        """
        nix_store_dir = self.env.dirs.real_store_dir
        state_dir = self.env.dirs.nix_state_dir

        # Make store writable
        Command(["chmod", "-R", "+w", nix_store_dir], self.env).run().ok()
        shutil.rmtree(nix_store_dir)
        shutil.rmtree(state_dir)

        # Re-create the directories
        nix_store_dir.mkdir()
        state_dir.mkdir()


_fully_sandboxed = (
    sys.platform == "linux"
    and Path("/proc/self/ns/user").is_symlink()
    and subprocess.run(["unshare", "--user", "--mount", "--pid", "true"]).returncode == 0
)


def pytest_runtest_setup(item: Any):
    for mark in item.iter_markers(name="full_sandbox"):
        if not _fully_sandboxed:
            pytest.skip(f"{sys.platform} does not support full sandboxing")


@pytest.fixture
def nix(tmp_path: Path, env: ManagedEnv, logger: logging.Logger) -> Generator[Nix, Any, None]:
    """
    Provides a rich way of calling `nix`.
    For pre-applied commands use `nix.nix_instantiate`, `nix.nix_build` etc.
    After configuring the command, use `.run()` to run it
    """
    yield Nix(env, logger)
    # when things are done using the nix store, the permissions for the store are read only
    # after the test was executed, we set the permissions to rwx (write being the important part)
    # for pytest to be able to delete the files during cleanup
    cmd = Command(argv=["chmod", "-R", "+w", str(tmp_path.absolute())], _env=env)
    cmd.run().ok()


@pytest.fixture
def enable_diverted_store(nix: Nix):
    """
    clear NIX_STORE_DIR, resetting it to the default (ie /nix/store).
    this makes builds impossible on platforms that cannot bind-mount,
    (e.g. macos) but it is important for eval result reproducibility.
    while builds may not work, substitution should still be possible.
    """
    nix.env.dirs.nix_store_dir = None
    nix.settings.sandbox_build_dir = None


def with_diverted_store(func: Callable[[Any], None]) -> Callable[[Any], None]:
    return pytest.mark.usefixtures("enable_diverted_store")(func)



import shutil
import pytest
import re
import dataclasses

from testlib.fixtures.command import CommandResult
from testlib.fixtures.nix import Nix
from testlib.fixtures.env import ManagedEnv


def build(nix: Nix, *args) -> CommandResult:
    expr = """
        derivation {
            name = "test";
            system = builtins.currentSystem;
            builder = "/bin/sh";
            args = [ "-c" "echo > $out" ];
            outputHashMode = "flat";
            outputHash = "sha256-AbpHGcgLb+kRsJGnwFEktk7uzpZOCcBY74+YBdrKVGs=";
        }
    """
    return nix.nix_build(["-E", expr, "--no-link", "--no-require-sigs", *args]).run()


@dataclasses.dataclass
class Caches:
    good: str
    bad: str


@pytest.fixture
def caches(nix: Nix, env: ManagedEnv) -> Caches:
    good_path, bad_path = env.dirs.home / "good", env.dirs.home / "bad"
    good_uri, bad_uri = f"file://{good_path}", f"file://{bad_path}"

    # build the derivation
    output = build(nix).ok().stdout_s.strip()
    # copy it to the good cache
    nix.nix(["copy", output, "--to", good_uri, "--no-require-sigs"], flake=True).run().ok()
    # create the bad cache by simulating read failures
    shutil.copytree(good_path, bad_path)
    for info in bad_path.glob("*.narinfo"):
        info.chmod(0o200)

    nix.nix(["store", "delete", output], flake=True).run().ok()

    return Caches(good=good_uri, bad=bad_uri)


def test_substitution_fallback_good_first(nix: Nix, caches: Caches):
    build(nix, "--substituters", f"{caches.good} {caches.bad}").ok()


def test_substitution_fallback_bad_first(nix: Nix, caches: Caches):
    # we expect three warnings for the single nar: two from querying, one from the substitution itself
    result = build(nix, "--substituters", f"{caches.bad} {caches.good}").ok()
    assert len(re.findall(r"warning.*narinfo", result.stderr_s)) == 3


def test_substitution_fallback_may_build(nix: Nix, caches: Caches):
    # we expect two errors for the single nar: one from querying, one from the substitution itself
    result = build(nix, "--substituters", f"{caches.bad}", "--fallback").ok()
    assert len(re.findall(r"error.*narinfo", result.stderr_s)) == 2


def test_substitution_fallback_no_build(nix: Nix, caches: Caches):
    # we expect one error, and it's fatal
    result = build(nix, "--substituters", f"{caches.bad}").expect(1)
    assert len(re.findall(r"error.*narinfo", result.stderr_s)) == 1



import os
import sys
import pytest
from testlib.fixtures.nix import Nix


@pytest.mark.skipif(
    sys.platform != "linux" or os.uname().machine != "x86_64", reason="uarch levels are x86_64 only"
)
def test_compute_levels(nix: Nix):
    assert "x86_64-v1-linux" in nix.nix(["-vv", "--version"]).run().ok().stdout_plain



from collections.abc import Callable
from pathlib import Path

from lang.test_lang import test_eval_okay as nix_eval
from testlib.fixtures.file_helper import AssetSymlink, CopyFile, CopyTree, with_files
from testlib.fixtures.nix import Nix
from testlib.fixtures.snapshot import Snapshot


@with_files(
    {
        "dir1": CopyTree("dir1"),
        "dir2": CopyTree("dir2"),
        "dir3": CopyTree("dir3"),
        "dir4": CopyTree("dir4"),
        "lib.nix": CopyFile("../lib.nix"),
        "in.nix": CopyFile("in.nix"),
        "out.exp": AssetSymlink("eval-okay.out.exp"),
        "err.exp": AssetSymlink("eval-okay.err.exp"),
    }
)
def test_search_path(files: Path, nix: Nix, snapshot: Callable[[str], Snapshot]):
    nix.env.set_env("NIX_PATH", "dir3:dir4")
    nix_eval(
        files,
        nix,
        [
            "--extra-deprecated-features",
            "shadow-internal-symbols",
            "-I",
            "dir1",
            "-I",
            "dir2",
            "-I",
            "dir5=dir3",
        ],
        snapshot,
    )


@with_files(
    {
        "nix-shadow": CopyTree("nix-shadow"),
        "in.nix": CopyFile("in-fetchurl.nix"),
        "out.exp": AssetSymlink("eval-okay-prefixed.out.exp"),
        "err.exp": AssetSymlink("eval-okay-prefixed.err.exp"),
    }
)
def test_prefixed_search_path(files: Path, nix: Nix, snapshot: Callable[[str], Snapshot]):
    nix.env.set_env("NIX_PATH", "nix=nix-shadow")
    nix_eval(files, nix, [], snapshot)


@with_files(
    {
        "nix-shadow": CopyTree("nix-shadow"),
        "in.nix": CopyFile("in-fetchurl.nix"),
        "out.exp": AssetSymlink("eval-okay-prefixless.out.exp"),
        "err.exp": AssetSymlink("eval-okay-prefixless.err.exp"),
    }
)
def test_prefixless_search_path(files: Path, nix: Nix, snapshot: Callable[[str], Snapshot]):
    nix_eval(files, nix, ["-I", "nix-shadow"], snapshot)


@with_files(
    {
        "in.nix": CopyFile("in-fetchurl.nix"),
        "out.exp": AssetSymlink("eval-okay-fetchurl.out.exp"),
        "err.exp": AssetSymlink("eval-okay-fetchurl.err.exp"),
    }
)
def test_empty_search_path(files: Path, nix: Nix, snapshot: Callable[[str], Snapshot]):
    nix_eval(files, nix, [], snapshot)



import re
from pathlib import Path

import pytest

from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack


@pytest.fixture(autouse=True)
def impure_vars(env: ManagedEnv):
    """
    Sets the IMPURE_VARs required by fixed.builder.sh
    because I have no clue what Eelco was thinking when writing this stuff
    """
    env["IMPURE_VAR1"] = "foo"
    env["IMPURE_VAR2"] = "bar"


@with_files(get_global_asset_pack("fixed"))
def test_bad(nix: Nix):
    res = nix.nix_instantiate(["fixed.nix", "-A", "good.0"]).run().ok()
    store_path = nix.nix_store(["-q", res.stdout_plain]).run().ok().stdout_plain
    path = Path(store_path)
    assert not path.exists()

    # Building with the bad hash should produce the "good" output path as
    # a sideeffect.
    res = nix.nix_build(["fixed.nix", "-A", "bad", "--no-out-link"]).run().expect(102)
    assert "hash mismatch in fixed-output derivation" in res.stderr_plain
    assert path.exists()

    nix.settings.add_xp_feature("nix-command")
    res = nix.nix(["path-info", "--json", store_path]).run().ok()
    assert res.json()[0]["ca"] == "fixed:md5:2qk15sxzzjlnpjk9brn7j8ppcd"


@with_files(get_global_asset_pack("fixed"))
def test_good(nix: Nix):
    nix.nix_build(["fixed.nix", "-A", "good", "--no-out-link"]).run().ok()


@with_files(get_global_asset_pack("fixed"))
def test_check(nix: Nix):
    res = nix.nix_build(["fixed.nix", "-A", "check", "--check"]).run().expect(1)
    assert "has no valid outputs registered in the store" in res.stderr_plain


@with_files(get_global_asset_pack("fixed"))
def test_good2(nix: Nix):
    # we need to build good.0 *first*, otherwise the hash good2 wants to be materialized
    # to does not exist. if it does not exist good2 would be built and *fail* since we'd
    # build a non-flat-hashable output. this is mondo jacked, but what can we do. -sigh-
    nix.nix_build(["fixed.nix", "-A", "good.0", "--no-out-link"]).run().ok()
    nix.nix_build(["fixed.nix", "-A", "good2", "--no-out-link"]).run().ok()


@with_files(get_global_asset_pack("fixed"))
def test_really_bad(nix: Nix):
    res = nix.nix_instantiate(["fixed.nix", "-A", "reallyBad"]).run().expect(1)
    assert (
        "error: hash 'ddd8be4b179a529afa5f2ffae4b9858' has wrong length for hash type 'md5'"
        in res.stderr_plain
    )


@with_files(get_global_asset_pack("fixed"))
def test_other_store_references(nix: Nix):
    res = nix.nix_build(["fixed.nix", "-A", "badReferences"]).run().expect(1)
    assert "not allowed to refer to other store paths" in res.stderr_plain


@with_files(get_global_asset_pack("fixed"))
def test_illegal_references(nix: Nix):
    """
    Fixed FOD hashes cannot be asserted because:
    - the store directory varies between the "Lix build sandbox environment" and a user test run
    - *-darwin has a different store location on the top of this in the sandbox (/private/tmp/...) causing further changes.
    Regex matching is the best we can afford.
    """
    res = nix.nix_build(["fixed.nix", "-A", "illegalReferences"]).run().expect(102)
    assert re.findall(
        r"the fixed-output derivation '.*?/nix/store/[a-z0-9]*-illegal-reference.drv' must not reference store paths but 1 such references were found:.*/nix/store/[a-z0-9]*-fixed",
        res.stderr_plain,
        re.DOTALL,
    )


@with_files(get_global_asset_pack("fixed"))
def test_attribute_selection(nix: Nix):
    nix.nix_instantiate(["fixed.nix", "-A", "good.1"]).run().ok()


@with_files(get_global_asset_pack("fixed"))
def test_parallel_same(nix: Nix):
    """
    Test parallel builds of derivations that produce the same output.
    Only one should run at the same time.
    """
    nix.nix_build(["fixed.nix", "-A", "parallelSame", "--no-out-link", "-j2"]).run().ok()


@pytest.mark.skip("TODO(rootile, 2025-10): Doesn't work for some reason, f1 test is kept for now")
@with_files(get_global_asset_pack("fixed"))
def test_same_as_add(nix: Nix, files: Path):
    """
    Fixed-output derivations with a recursive SHA-256 hash should
    produce the same path as "nix-store --add".
    """
    res = nix.nix_build(["fixed.nix", "-A", "sameAsAdd", "--no-out-link"]).run().ok()
    out = res.stdout_plain

    fixed = files / "fixed"
    # is failing as this isn't created for f2: shutil.rmtree(fixed)
    (fixed / "bla").mkdir(parents=True)
    foo = fixed / "foo"
    foo.write_text("Hello World!")
    (fixed / "bar").symlink_to(foo)

    out2 = nix.nix_store(["--add", str(fixed)]).run().ok().stdout_plain
    assert out == out2

    out3 = (
        nix.nix_store(["--add-fixed", "--recursive", "sha256", str(fixed)]).run().ok().stdout_plain
    )
    assert out == out3

    out4 = (
        nix.nix_store(
            [
                "--print-fixed-path",
                "--recursive",
                "sha256",
                "1ixr6yd3297ciyp9im522dfxpqbkhcw0pylkb2aab915278fqaik",
                "fixed",
            ]
        )
        .run()
        .ok()
        .stdout_plain
    )
    assert out == out4



from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack


@with_files(get_global_asset_pack("dependencies"))
def test_command_doesnt_crash(nix: Nix):
    nix.nix_store(["--init"]).run().ok()
    nix.nix(
        ["why-depends", "--derivation", "--file", "./dependencies.nix", "input2_drv", "input1_drv"],
        flake=True,
    ).run().ok()
    nix.nix(
        ["why-depends", "--file", "./dependencies.nix", "input2_drv", "input1_drv"], flake=True
    ).run().ok()


def built_files(nix: Nix):
    nix.nix_build(["./dependencies.nix", "-A", "input0_drv", "-o", "dep"]).run().ok()
    nix.nix_build(["./dependencies.nix", "-A", "input3_drv", "-o", "dep3"]).run().ok()
    nix.nix_build(["./dependencies.nix", "-o", "toplevel"]).run().ok()


@with_files(get_global_asset_pack("dependencies"))
def test_fast_depends(nix: Nix):
    built_files(nix)
    res = nix.nix(["why-depends", "./toplevel", "./dep"], flake=True).run().ok()
    assert "input-2" in res.stdout_plain
    assert "reference-to-input-2" not in res.stdout_plain


@with_files(get_global_asset_pack("dependencies"))
def test_precise_depends(nix: Nix):
    built_files(nix)
    res = nix.nix(["why-depends", "./toplevel", "./dep", "--precise"], flake=True).run().ok()
    out = res.stdout_plain
    assert "input-2" in out
    assert "reference-to-input-2" in out
    lines = out.splitlines()
    assert "reference-to-input-2 -> " in lines[1]
    assert "dependencies-input-2" in lines[2]
    assert "input0: " in lines[3]
    assert "dependencies-input-0" in lines[4]


@with_files(get_global_asset_pack("dependencies"))
def test_self_ref(nix: Nix):
    built_files(nix)
    res = nix.nix(["why-depends", "./toplevel", "./toplevel"], flake=True).run().ok()
    lines = res.stdout_plain.splitlines()
    assert "dependencies-top" in lines[0]
    assert "dependencies-top" in lines[1]


@with_files(get_global_asset_pack("dependencies"))
def test_all(nix: Nix):
    built_files(nix)
    res = nix.nix(["why-depends", "./toplevel", "./dep3", "--all"], flake=True).run().ok()
    lines = res.stdout_plain.splitlines()
    assert "dependencies-top" in lines[0]
    assert "input-2" in lines[1]
    assert "input-3" in lines[2]



import re

from testlib.fixtures.file_helper import with_files, CopyFile
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset

_fod_files = {
    "fod-failing.nix": CopyFile("assets/test_build/fod-failing.nix"),
    "config.nix": get_global_asset("config.nix"),
}
_build_args = ["build", "-f", "fod-failing.nix", "-L"]


@with_files(_fod_files)
def test_url_mismatch(nix: Nix):
    res = nix.nix_build(["fod-failing.nix", "-A", "x1"]).run().expect(102)
    err = res.stderr_plain
    assert len(re.findall(r"hash mismatch in fixed-output derivation '.*-x1\.drv'", err)) == 1
    assert "likely URL: https://meow.puppy.forge/puppy.tar.gz" in err


@with_files(_fod_files)
def test_url_keep_going_1(nix: Nix):
    res = nix.nix([*_build_args], flake=True).run().expect(1)
    err = res.stderr_plain
    assert err.count("error:") >= 2
    assert re.findall(r"hash mismatch in fixed-output derivation '.*-x.\.drv'", err)
    assert "likely URL: " in err
    assert re.findall(
        r"error: build of '.*-x[1-4]\.drv\^out', '.*-x[1-4]\.drv\^out', '.*-x[1-4]\.drv\^out', '.*-x[1-4]\.drv\^out' failed",
        err,
    )


@with_files(_fod_files)
def test_url_keep_going_2(nix: Nix):
    res = nix.nix([*_build_args, "x1", "x2", "x3", "--keep-going"], flake=True).run().expect(1)
    err = res.stderr_plain
    assert err.count("error:") == 4
    for f in ["x1", "x2", "x3"]:
        assert re.findall(rf"hash mismatch in fixed-output derivation '.*-{f}\.drv'", err)
    assert "likely URL: https://meow.puppy.forge/puppy.tar.gz" in err
    assert "likely URL: https://kitty.forge/cat.tar.gz" in err
    assert "likely URL: (unknown)" in err
    assert re.findall(
        r"error: build of '.*-x[1-3]\.drv\^out', '.*-x[1-3]\.drv\^out', '.*-x[1-3]\.drv\^out' failed",
        err,
    )


@with_files(_fod_files)
def test_missing_dependency(nix: Nix):
    res = nix.nix([*_build_args, "x4"], flake=True).run().expect(1)
    err = res.stderr_plain
    assert err.count("error:") >= 2
    assert re.findall(r"error: [12] dependencies of derivation '.*-x4\.drv' failed to build", err)
    assert re.findall(r"hash mismatch in fixed-output derivation '.*-x[23]\.drv'", err)


@with_files(_fod_files)
def test_missing_dependency_keep_going(nix: Nix):
    res = nix.nix([*_build_args, "x4", "--keep-going"], flake=True).run().expect(1)
    err = res.stderr_plain
    assert err.count("\nerror:") == 3
    assert re.findall(r"error: 2 dependencies of derivation '.*-x4\.drv' failed to build", err)
    for f in ["x2", "x3"]:
        assert re.findall(rf"hash mismatch in fixed-output derivation '.*-{f}\.drv'", err)


@with_files({"config.nix": get_global_asset("config.nix")})
def test_build_inaccessible_build_dir(nix: Nix):
    env = nix.env
    new_build_dir = env.dirs.home / "build-dir"
    new_build_dir.mkdir()
    new_build_dir.chmod(0o0000)
    try:
        nix.nix(
            [
                "--build-dir",
                str(new_build_dir),
                "build",
                "-E",
                'with import ./config.nix; mkDerivation { name = "test"; buildCommand = "echo rawr > $out"; }',
                "--impure",
                "--no-link",
            ],
            flake=True,
        ).run().ok()
    finally:
        # clean up perms
        new_build_dir.chmod(0o777)



#!/usr/bin/env python3

"""
So like. This script is cursed.
It's a helper for this project's Meson buildsystem for Lix's functional tests.
The functional tests are a bunch of bash scripts, that each expect to be run from the
directory from the directory that that script is in, and also expect modifications to have
happened to the source tree, and even splork files around. The last is against the spirit
of Meson (and personally annoying), to have build processes that aren't self-contained in the
out-of-source build directory, but more problematically we need configured files in the test
tree.
So. We copy the tests tree into the build directory.
Meson doesn't have a good way of doing this natively -- the best you could do is subdir()
into every directory in the tests tree and configure_file(copy : true) on every file,
but this won't copy symlinks as symlinks, which we need since the test suite has, well,
tests on symlinks.
However, the functional tests are normally run during Lix's derivation's installCheckPhase,
after Lix has already been "installed" somewhere. So in Meson we setup add this file as an
install script and copy everything in tests/functional to the build directory, preserving
things like symlinks, even broken ones (which are intentional).

TODO(Qyriad): when we remove the old build system entirely, we can instead fix the tests.
"""

from pathlib import Path
import os, os.path
import shutil
import sys
import traceback

name = 'setup-functional-tests.py'

if 'MESON_SOURCE_ROOT' not in os.environ or 'MESON_BUILD_ROOT' not in os.environ:
    raise ValueError(f'{name}: this script must be run from the Meson build system')

print(f'{name}: mirroring tests/functional to build directory')

tests_source = Path(os.environ['MESON_SOURCE_ROOT']) / 'tests/functional'
tests_build = Path(os.environ['MESON_BUILD_ROOT']) / 'tests/functional'

def main():

    os.chdir(tests_build)

    for src_dirpath, src_dirnames, src_filenames in os.walk(tests_source):
        src_dirpath = Path(src_dirpath)
        assert src_dirpath.is_absolute(), f'{src_dirpath=} is not absolute'

        # os.walk() gives us the absolute path to the directory we're currently in as src_dirpath.
        # We want to mirror from the perspective of `tests_source`.
        rel_subdir = src_dirpath.relative_to(tests_source)
        assert (not rel_subdir.is_absolute()), f'{rel_subdir=} is not relative'

        # And then join that relative path on `tests_build` to get the absolute
        # path in the build directory that corresponds to `src_dirpath`.
        build_dirpath = tests_build / rel_subdir
        assert build_dirpath.is_absolute(), f'{build_dirpath=} is not absolute'

        # More concretely, for the test file tests/functional/ca/build.sh:
        # - src_dirpath is `$MESON_SOURCE_ROOT/tests/functional/ca`
        # - rel_subidr is `ca`
        # - build_dirpath is `$MESON_BUILD_ROOT/tests/functional/ca`

        # `src_dirname` are directories underneath `src_dirpath`, and will be relative
        # to `src_dirpath`.
        for src_dirname in src_dirnames:
            # Take the name of the directory in the tests source and join it on `src_dirpath`
            # to get the full path to this specific directory in the tests source.
            src = src_dirpath / src_dirname
            # If there isn't *something* here, then our logic is wrong.
            # Path.exists(follow_symlinks=False) wasn't added until Python 3.12, so we use
            # os.path.lexists() here.
            assert os.path.lexists(src), f'{src=} does not exist'

            # Take the name of this directory and join it on `build_dirpath` to get the full
            # path to the directory in `build/tests/functional` that we need to create.
            dest = build_dirpath / src_dirname
            if src.is_symlink():
                src_target = src.readlink()
                dest.unlink(missing_ok=True)
                dest.symlink_to(src_target)
            else:
                dest.mkdir(parents=True, exist_ok=True)

        for src_filename in src_filenames:
            # os.walk() should be giving us relative filenames.
            # If it isn't, our path joins will be veeeery wrong.
            assert (not Path(src_filename).is_absolute()), f'{src_filename=} is not relative'

            src = src_dirpath / src_filename
            dst = build_dirpath / src_filename
            # Mildly misleading name -- unlink removes ordinary files as well as symlinks.
            dst.unlink(missing_ok=True)
            # shutil.copy2() best-effort preserves metadata.
            shutil.copy2(src, dst, follow_symlinks=False)


try:
    sys.exit(main())
except Exception as e:
    # Any error is likely a bug in this script.
    print(f'{name}: INTERNAL ERROR setting up functional tests: {e}', file=sys.stderr)
    print(traceback.format_exc())
    print(f'this is a bug in {name}')
    sys.exit(1)



import pytest

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack


@with_files(get_global_asset_pack("dependencies"))
@pytest.mark.parametrize("algorithm", ["br", "zstd", "xz"])
def test_cache_compressions(nix: Nix, algorithm: str):
    nix.settings.add_xp_feature("nix-command")
    cache_uri = f"file://{nix.env.dirs.cache_dir}?compression={algorithm}"
    res = nix.nix_build(["dependencies.nix", "--no-out-link"]).run().ok()
    out_path = res.stdout_plain

    res = nix.nix(["copy", "--to", cache_uri, out_path]).run().ok()
    assert "copying 4 paths..." in res.stderr_plain
    hash1 = nix.hash_path(out_path)

    # clear the "cache cache"
    for f in (nix.env.dirs.home / ".cache" / "nix").glob("test-binary-cache*"):
        f.unlink()
    nix.clear_store()

    res = nix.nix(["copy", "--from", cache_uri, out_path, "--no-check-sigs"]).run().ok()
    assert "copying 4 paths..." in res.stderr_plain

    hash2 = nix.hash_path(out_path)

    assert hash1 == hash2



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File, EnvTemplate
from testlib.utils import get_global_asset_pack
from pathlib import Path
import pytest
import fnmatch
import json


@pytest.fixture(autouse=True)
def add_xp_features(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")


@with_files(
    {
        "flakeA": get_global_asset_pack(".git")
        | {
            "flake.nix": EnvTemplate("""{
                inputs = {
                    B = {
                        url = "path:./flakeB";
                        inputs.foobar.follows = "foobar";
                    };
                    foobar.url = "path:@HOME@/flakeA/flakeE";
                };
                outputs = { ... }: {};
            }"""),
            "flakeB": {
                "flake.nix": EnvTemplate("""{
                    inputs = {
                        foobar.url = "path:@HOME@/flakeA/flakeE";
                        goodoo.follows = "C/goodoo";
                        C = {
                            url = "path:./flakeC";
                            inputs.foobar.follows = "foobar";
                        };
                    };
                    outputs = { ... }: {};
                }"""),
                "flakeC": {
                    "flake.nix": EnvTemplate("""{
                        inputs = {
                            foobar.url = "path:@HOME@/flakeA/flakeE";
                            goodoo.follows = "foobar";
                        };
                        outputs = { ... }: {};
                    }""")
                },
            },
            "flakeD": {"flake.nix": File("{ outputs = { ... }: {}; }")},
            "flakeE": {"flake.nix": File("{ outputs = { ... }: {}; }")},
        }
    }
)
class TestPathFollows:
    @pytest.fixture(autouse=True)
    def init(self, git: Git, files: Path):
        git(files / "flakeA", "add", ".")

    def test_metadata(self, nix: Nix, files: Path):
        out = nix.nix(["flake", "metadata", files / "flakeA"]).run().ok()
        assert "B: path" in out.stdout_s
        assert "C: path" in out.stdout_s
        assert "foobar: path" in out.stdout_s
        assert "goodoo follows" in out.stdout_s

    def test_update(self, nix: Nix, files: Path):
        out = nix.nix(["flake", "update", "--flake", files / "flakeA"]).run().ok()
        assert "Added input 'B" in out.stderr_s
        assert "Added input 'B/C" in out.stderr_s
        assert "Added input 'B/C/foobar" in out.stderr_s
        assert "Added input 'B/C/goodoo" in out.stderr_s
        assert "Added input 'B/foobar" in out.stderr_s
        assert "Added input 'B/goodoo" in out.stderr_s

    def test_lock(self, nix: Nix, files: Path):
        nix.nix(["flake", "lock", files / "flakeA"]).run().ok()
        lockfile = json.loads((files / "flakeA/flake.lock").read_text())

        assert lockfile["nodes"]["B"]["inputs"]["C"] == "C"
        assert lockfile["nodes"]["B"]["inputs"]["foobar"] == ["foobar"]
        assert lockfile["nodes"]["C"]["inputs"]["foobar"] == ["B", "foobar"]

        # Ensure that locking twice doesn't change anything
        nix.nix(["flake", "lock", files / "flakeA"]).run().ok()
        lockfile2 = json.loads((files / "flakeA/flake.lock").read_text())
        assert lockfile == lockfile2

    def test_removing_follows_updates_lockfile(self, nix: Nix, files: Path, git: Git):
        git(files / "flakeA", "add", ".")

        nix.nix(["flake", "lock", files / "flakeA"]).run().ok()
        (files / "flakeA/flake.nix").write_text("""{
            inputs = {
                B.url = "path:./flakeB";
                D.url = "path:./flakeD";
            };
            outputs = { ... }: {};
        }""")

        nix.nix(["flake", "lock", files / "flakeA"]).run().ok()

        lockfile = json.loads((files / "flakeA/flake.lock").read_text())
        assert lockfile["nodes"]["B"]["inputs"]["foobar"] == "foobar"
        assert "foobar" in lockfile["nodes"]


@with_files(
    {
        "flakeA": get_global_asset_pack(".git")
        | {
            "flake.nix": File("""{
                inputs.B.url = "path:../flakeB";
                outputs = { ... }: {};
            }""")
        }
    }
)
def test_relative_path_cant_leave_store_path(nix: Nix, files: Path, git: Git):
    git(files / "flakeA", "add", ".")

    out = nix.nix(["flake", "lock", files / "flakeA"]).run().expect(1)
    assert "points outside" in out.stderr_s


@with_files(
    {
        "flakeA": get_global_asset_pack(".git")
        | {
            "flake.nix": File("""{
                inputs.B = {
                    url = "path:./flakeB";
                    inputs.invalid.follows = "D";
                    inputs.invalid2.url = "path:./flakeD";
                };
                inputs.D.url = "path:./flakeD";
                outputs = { ... }: {};
            }"""),
            "flakeB": {
                "flake.nix": EnvTemplate("""{
                    inputs = {
                        foobar.url = "path:@HOME@/flakeA/flakeE";
                        goodoo.follows = "C/goodoo";
                        C = {
                            url = "path:./flakeC";
                            inputs.foobar.follows = "foobar";
                        };
                    };
                    outputs = { ... }: {};
                }"""),
                "flakeC": {
                    "flake.nix": EnvTemplate("""{
                        inputs = {
                            foobar.url = "path:@HOME@/flakeA/flakeE";
                            goodoo.follows = "foobar";
                        };
                        outputs = { ... }: {};
                    }""")
                },
            },
            "flakeD": {"flake.nix": File("{ outputs = { ... }: {}; }")},
            "flakeE": {"flake.nix": File("{ outputs = { ... }: {}; }")},
        }
    }
)
def test_nonexistent_follows_print_warnings(nix: Nix, files: Path, git: Git):
    git(files / "flakeA", "add", ".")

    out = nix.nix(["flake", "lock", files / "flakeA"]).run().ok()
    assert "warning: input 'B' has an override for a non-existent input 'invalid'" in out.stderr_s
    assert "warning: input 'B' has an override for a non-existent input 'invalid2'" in out.stderr_s


@with_files(
    {
        "flakeA": get_global_asset_pack(".git")
        | {
            # input B/D should be able to be found...
            "flake.nix": File("""{
                inputs = {
                    B = {
                        url = "path:./flakeB";
                        inputs.C.follows = "C";
                    };
                    C.url = "path:./flakeB/flakeC";
                };
                outputs = { ... }: {};
            }"""),
            "flakeB": {
                "flake.nix": File("""{
                    inputs = {
                        C.url = "path:./flakeC";
                        D.follows = "C/D";
                    };
                    outputs = { ... }: {};
                }"""),
                "flakeC": {
                    "flake.nix": File("""{
                        inputs.D.url = "path:./flakeD";
                        outputs = { ... }: {};
                    }"""),
                    "flakeD": {"flake.nix": File("{ outputs = { ... }: {}; }")},
                },
            },
        }
    }
)
def test_follow_path_overloading(nix: Nix, files: Path, git: Git):
    r"""
    This tests a lockfile checking regression https://github.com/NixOS/nix/pull/8819

    We construct the following graph, where p->q means p has input q.
    A double edge means that the edge gets overridden using `follows`.

         A
        / \
       /   \
      v     v
      B ==> C   --- follows declared in A
       \\  /
        \\/     --- follows declared in B
         v
         D

    The message was
       error: input 'B/D' follows a non-existent input 'B/C/D'

    Note that for `B` to resolve its follow for `D`, it needs `C/D`,
    for which it needs to resolve the follow on `C` first.
    """

    git(files / "flakeA", "add", ".")

    nix.nix(["flake", "metadata", f"{files}/flakeA"]).run().ok()
    nix.nix(["flake", "update", "--flake", f"{files}/flakeA"]).run().ok()
    nix.nix(["flake", "lock", f"{files}/flakeA"]).run().ok()

    lockfile = json.loads((files / "flakeA/flake.lock").read_text())
    assert lockfile["nodes"]["B"]["inputs"] == {"C": ["C"], "D": ["B", "C", "D"]}
    assert lockfile["nodes"]["C"]["inputs"] == {"D": "D"}
    assert lockfile["nodes"]["root"]["inputs"] == {"B": "B", "C": "C"}


@with_files(
    {
        "flakeA": {
            "flake.nix": EnvTemplate("""{
                inputs.B.url = "path:@HOME@/flakeA/flakeB";
                inputs.D.url = "path:@HOME@/flakeA/flakeD";
                inputs.B.inputs.C.inputs.D.follows = "D";
                outputs = _: {};
            }"""),
            "flakeB": {
                "flake.nix": EnvTemplate("""{
                    inputs.C.url = "path:@HOME@/flakeA/flakeB/flakeC";
                    outputs = _: {};
                }"""),
                "flakeC": {
                    "flake.nix": EnvTemplate("""{
                      inputs.D.url = "path:nosuchflake";
                      outputs = _: {};
                    }""")
                },
            },
            "flakeD": {"flake.nix": File("{ outputs = _: {}; }")},
        }
    }
)
def test_nested_overrides(nix: Nix, files: Path):
    nix.nix(["flake", "lock", f"{files}/flakeA"]).run().ok()
    lockfile = json.loads((files / "flakeA/flake.lock").read_text())
    assert lockfile["nodes"]["C"]["inputs"]["D"] == ["D"]


@with_files(
    {
        "flakeA": {
            "flake.nix": EnvTemplate("""{
                inputs.B.url = "path:@HOME@/flakeA/flakeB";
                inputs.C.url = "path:@HOME@/flakeA/flakeB/flakeC";
                inputs.B.inputs.C.follows = "C";
                outputs = _: {};
            }"""),
            "flakeB": {
                "flake.nix": EnvTemplate("""{
                    inputs.C.url = "path:nosuchflake";
                    inputs.D.url = "path:nosuchflake";
                    inputs.D.follows = "C/D";
                    outputs = _: {};
                }"""),
                "flakeC": {
                    "flake.nix": EnvTemplate("""{
                      inputs.D.url = "path:@HOME@/flakeA/flakeD";
                      outputs = _: {};
                    }""")
                },
            },
            "flakeD": {"flake.nix": File("{ outputs = _: {}; }")},
        }
    }
)
def test_overlapping_flake_follows(nix: Nix, files: Path):
    # bug was not triggered without recreating the lockfile
    nix.nix(["flake", "update", "--flake", f"{files}/flakeA"]).run().ok()
    lockfile = json.loads((files / "flakeA/flake.lock").read_text())
    assert lockfile["nodes"]["B"]["inputs"]["D"] == ["B", "C", "D"]


@with_files(
    {
        "flakeA": {
            "flake.nix": EnvTemplate("""{
                inputs.B.url = "path:@HOME@/flakeA/flakeB";
                inputs.C.url = "path:@HOME@/flakeA/flakeB/flakeC";
                inputs.B.inputs.C.inputs.E.follows = "C";
                outputs = _: {};
            }"""),
            "flakeB": {
                "flake.nix": EnvTemplate("""{
                    inputs.C.url = "path:@HOME@/flakeA/flakeB/flakeC";
                    outputs = _: {};
                }"""),
                "flakeC": {
                    "flake.nix": EnvTemplate("""{
                      inputs.D.url = "path:@HOME@/flakeA/flakeD";
                      outputs = _: {};
                    }""")
                },
            },
            "flakeD": {"flake.nix": File("{ outputs = _: {}; }")},
        }
    }
)
def test_override_nonexistant_input(nix: Nix, files: Path):
    """Test overlapping flake follows: B has D follow C/D, while A has B/C follow C"""
    out = nix.nix(["flake", "update", "--flake", f"{files}/flakeA"]).run().ok()
    assert "warning: input 'B/C' has an override for a non-existent input 'E'" in out.stderr_s


@with_files(
    {
        "flakeA": {
            "flake.nix": EnvTemplate("""{
                inputs = {
                  B.url = "path:@HOME@/flakeA/flakeB";
                  C = {
                    url = "path:@HOME@/flakeA/flakeB/flakeC";
                    inputs.D.inputs.E.follows = "B";
                  };
                };
                outputs = _: {};
            }"""),
            "flakeB": {
                "flake.nix": File("{ outputs = _: {}; }"),
                "flakeC": {
                    "flake.nix": EnvTemplate("""{
                      inputs.D.url = "path:@HOME@/flakeA/flakeD";
                      outputs = _: {};
                    }""")
                },
            },
            "flakeD": {
                "flake.nix": File("""{
                    inputs.E.url = "path:@HOME@/flakeA/flakeE";
                    outputs = _: {};
                }""")
            },
            "flakeE": {"flake.nix": File("{ outputs = { ... }: {}; }")},
        }
    }
)
def test_dep_fetch(nix: Nix, files: Path):
    """Test for Nested follows cause flake interactions to update the nested input #460"""

    # Lockfiles are cleared, initially the dependency needs to be fetched.
    out = nix.nix(["--verbose", "flake", "show", f"path:{files}/flakeA"]).run().ok()
    assert fnmatch.fnmatch(out.stderr_s, "*\nfetching path input 'path:*/flakeD'\n*")

    # But on another flake command it doesn't.
    out = nix.nix(["--verbose", "flake", "show", f"path:{files}/flakeA"]).run().ok()
    assert not fnmatch.fnmatch(out.stderr_s, "*\nfetching path input 'path:*/flakeD'\n*")

    # Make sure the nested override is actually correct in this testcase.
    metadata = nix.nix(["flake", "metadata", f"path:{files}/flakeA", "--json"]).run().json()
    assert metadata["locks"]["nodes"]["D"]["inputs"]["E"] == ["B"]



import logging
from collections.abc import Callable
from pathlib import Path
from textwrap import dedent

import pytest
from _pytest.logging import LogCaptureFixture
from testlib.fixtures.command import Command
from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.file_helper import (
    CopyFile,
    File,
    FileDeclaration,
    merge_file_declaration,
    with_files,
)
from testlib.fixtures.snapshot import Snapshot
from testlib.utils import get_functional2_files


def _get_f2_snapshot_files(additional_files: FileDeclaration) -> FileDeclaration:
    return get_functional2_files(
        merge_file_declaration(
            {
                "functional2": {
                    "testlib": {
                        "__init__.py": CopyFile("../__init__.py"),
                        "environ.py": CopyFile("../environ.py"),
                        "fixtures": {
                            "__init__.py": CopyFile("__init__.py"),
                            "snapshot.py": CopyFile("snapshot.py"),
                            "logger.py": CopyFile("logger.py"),
                        },
                    },
                    "conftest.py": File(
                        "pytest_plugins = ('testlib.fixtures.logger', 'testlib.fixtures.snapshot')"
                    ),
                }
            },
            additional_files,
        )
    )


def test_snapshot_empty_on_no_file(snapshot: Callable[[str], Snapshot]):
    # no qa here, as we want to exactly check for empty string
    assert snapshot("this_file_does_not_exist") == ""  # noqa: PLC1901


@with_files({"empty_file.txt": File("")})
def test_snapshot_empty_on_empty_file(snapshot: Callable[[str], Snapshot]):
    # no qa here, as we want to exactly check for empty string
    assert snapshot("empty_file.txt") == ""  # noqa: PLC1901


@with_files({"out.exp": File("Hell o' World")})
def test_snapshot_warns_on_non_symlink_file(
    snapshot: Callable[[str], Snapshot], caplog: LogCaptureFixture
):
    with caplog.at_level(logging.WARNING):
        snapshot("out.exp")
    assert len(caplog.records) == 1
    assert (
        caplog.records[0].msg
        == "expected output file isn't a symlink. When accepting the tests output, the update might not propagate"
    )


@pytest.mark.parametrize("pytest_command", [([], False)], indirect=True)
@with_files(
    _get_f2_snapshot_files(
        {
            "functional2": {
                "test_snapshot": {
                    "test_snapshot.py": File(
                        dedent("""
                        def test_noupdate(do_snapshot_update):
                            assert not do_snapshot_update
                        """)
                    )
                }
            }
        }
    )
)
def test_do_update_false_when_none_set(pytest_command: Command):
    pytest_command.run().ok()


_update_test_files = _get_f2_snapshot_files(
    {
        "functional2": {
            "test_snapshot": {
                "test_snapshot.py": File(
                    dedent("""
                        def test_does_update(do_snapshot_update):
                            assert do_snapshot_update
                        """)
                )
            }
        }
    }
)


@pytest.mark.parametrize(
    "pytest_command", [([], False), (["--accept-tests"], False)], indirect=True
)
@pytest.mark.parametrize("set_env", [False, True])
@with_files(_update_test_files)
def test_do_update_true_when_any_set(env: ManagedEnv, pytest_command: Command, set_env: bool):
    if not (set_env or "--accept-tests" in pytest_command.argv):
        pytest_command.discard()
        pytest.skip("not this test case")
    if set_env:
        env.set_env("_NIX_TEST_ACCEPT", "1")
    pytest_command.run().ok()


def _snapshot_test_files(content: str) -> FileDeclaration:
    return _get_f2_snapshot_files(
        {
            "functional2": {
                "test_snapshot": {
                    "test_snapshot.py": File(
                        dedent(f"""
                            def test_snapshot(snapshot, tmp_path):
                                (tmp_path / "test-home" / "out.exp").write_text("{content}")
                                assert snapshot("out.exp") == "plush plush"
                            """)
                    )
                }
            }
        }
    )


@pytest.mark.parametrize("pytest_command", [(["--accept-tests"], False)], indirect=True)
@with_files(_snapshot_test_files("plush plush"))
def test_snapshot_no_updated_when_equal(pytest_command: Command):
    res = pytest_command.run().ok()
    assert "the updated file can be found here" not in res.stdout_plain


@pytest.mark.parametrize("pytest_command", [([], False)], indirect=True)
@with_files(_snapshot_test_files("fops plush"))
def test_snapshot_fails_on_diff(pytest_command: Command):
    res = pytest_command.run().expect(1)
    assert "FAILED test_snapshot/test_snapshot.py::test_snapshot" in res.stdout_plain


@pytest.mark.parametrize("pytest_command", [(["--accept-tests"], False)], indirect=True)
@with_files(_snapshot_test_files("fops plush"))
def test_snapshot_updates_diff(files: Path, pytest_command: Command):
    output_file = files / "pytest_files/test_snapshot0/test-home/out.exp"
    pytest_command.run().ok()
    assert output_file.read_text() == "plush plush"


@pytest.mark.parametrize("pytest_command", [(["--accept-tests"], False)], indirect=True)
@with_files(_snapshot_test_files("fops plush"))
def test_snapshot_updates_shares_updated_location_without_symlink(
    files: Path, pytest_command: Command
):
    expected_path = "pytest_files/test_snapshot0/test-home/out.exp"
    output_file = files / expected_path
    res = pytest_command.run().ok()
    assert output_file.read_text() == "plush plush"
    assert "the updated file can be found here" in res.stdout_plain
    assert expected_path in res.stdout_plain


@pytest.mark.parametrize("pytest_command", [(["--accept-tests"], False)], indirect=True)
@with_files(_snapshot_test_files("fops plush"))
def test_snapshot_marks_skip_after_update(files: Path, pytest_command: Command):
    expected_path = "pytest_files/test_snapshot0/test-home/out.exp"
    output_file = files / expected_path
    res = pytest_command.run().ok()
    assert output_file.read_text() == "plush plush"
    assert "test_snapshot/test_snapshot.py::test_snapshot SKIPPED (Updated" in res.stdout_plain


@pytest.mark.parametrize("pytest_command", [(["--accept-tests"], False)], indirect=True)
@with_files(
    _get_f2_snapshot_files(
        {
            "functional2": {
                "test_snapshot": {
                    "test_snapshot.py": File(
                        dedent("""
                        def test_snapshot(snapshot, tmp_path):
                            (tmp_path / "test-home" / "out.exp").write_text("fops plush")
                            (tmp_path / "test-home" / "err.exp").write_text("shork plush")
                            assert snapshot("out.exp") == "plush plush"
                            assert snapshot("err.exp") == "blobhaj"
                        """)
                    )
                }
            }
        }
    )
)
def test_snapshot_updates_multiple(files: Path, pytest_command: Command):
    expected_path = "pytest_files/test_snapshot0/test-home"
    first_file = files / expected_path / "out.exp"
    second_file = files / expected_path / "err.exp"
    res = pytest_command.run().ok()
    assert "test_snapshot/test_snapshot.py::test_snapshot SKIPPED (Updated" in res.stdout_plain
    assert first_file.read_text() == "plush plush"
    assert second_file.read_text() == "blobhaj"


@pytest.mark.parametrize("pytest_command", [(["--accept-tests"], False)], indirect=True)
@with_files(
    _get_f2_snapshot_files(
        {
            "functional2": {
                "test_snapshot": {
                    "test_snapshot.py": File(
                        dedent("""
                        def test_snapshot(snapshot, tmp_path):
                            (tmp_path / "test-home" / "updated.txt").write_text("snek plush")
                            (tmp_path / "test-home" / "out.exp").symlink_to("./updated.txt")
                            assert snapshot("out.exp") == "plush plush"
                        """)
                    )
                }
            }
        }
    )
)
def test_snapshot_updates_no_location_when_symlink(files: Path, pytest_command: Command):
    expected_path = "pytest_files/test_snapshot0/test-home/updated.txt"
    output_file = files / expected_path
    res = pytest_command.run().ok()
    assert output_file.read_text() == "plush plush"
    assert "the updated file can be found here" not in res.stdout_plain



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File, EnvTemplate, CopyFile
from testlib.environ import environ
from testlib.utils import get_global_asset, get_global_asset_pack
from pathlib import Path
import pytest

system = environ.get("system")

flake = {
    "flake.nix": EnvTemplate(f"""{{
        inputs.nixpkgs.url = "@HOME@/nixpkgs";
        outputs = {{self, nixpkgs}}: {{
          packages.{system}.hello = (import ./config.nix).mkDerivation {{
            name = "hello";
            outputs = [ "out" "dev" ];
            meta.outputsToInstall = [ "out" ];
            buildCommand = "";
          }};
        }};
    }}"""),
    "config.nix": get_global_asset("config.nix"),
    "shell-hello.nix": CopyFile("assets/shell-hello.nix"),
}
nixpkgs = {
    "flake.nix": File(f"""{{
        outputs = {{self}}: {{
          legacyPackages.{system}.bashInteractive = (import ./shell.nix {{}}).bashInteractive;
        }};
    }}"""),
    "config.nix": get_global_asset("config.nix"),
    "shell.nix": get_global_asset("shell.nix"),
}


@pytest.fixture(autouse=True)
def common_init(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")


@with_files(flake | {"nixpkgs": nixpkgs})
class TestDevelop:
    def test_env_passthrough(self, nix: Nix):
        nix.env["ENVVAR"] = "a"

        result = (
            nix.nix(["develop", "--no-write-lock-file", ".#hello"])
            .with_stdin(b'echo "$ENVVAR"')
            .run()
            .ok()
        )
        assert result.stdout_s == "a\n"

    def test_no_env_passthrough_on_ignore_environment(self, nix: Nix):
        nix.env["ENVVAR"] = "a"

        result = (
            nix.nix(["develop", "--ignore-environment", "--no-write-lock-file", ".#hello"])
            .with_stdin(b'echo "$ENVVAR"')
            .run()
            .ok()
        )
        assert result.stdout_s == "\n"

    class TestShell:
        @pytest.fixture(autouse=True)
        def shell_init(self, nix: Nix, files: Path):  # noqa: ARG002
            nix.nix(
                [
                    "build",
                    "--no-write-lock-file",
                    "./nixpkgs#bashInteractive",
                    "--out-link",
                    "./bash-interactive",
                ]
            ).run().ok()
            self.bash_interactive = nix.env.dirs.home / "bash-interactive/bin/bash"

            nix.env["SHELL"] = "custom"

        @pytest.mark.parametrize("extra_args", [[], ["--ignore-environment"]])
        def test_shell_is_bash_interactive(self, nix: Nix, extra_args: list[str]):
            result = (
                nix.nix(["develop", *extra_args, "--no-write-lock-file", ".#hello"])
                .with_stdin(b'echo "$SHELL"')
                .run()
                .ok()
            )
            assert Path(result.stdout_s.strip()).samefile(self.bash_interactive)

    @pytest.mark.parametrize("extra_args", [[], ["--ignore-environment"]])
    def test_sets_in_nix_shell(self, nix: Nix, extra_args: list[str]):
        result = (
            nix.nix(
                [
                    "develop",
                    "--no-write-lock-file",
                    *extra_args,
                    ".#hello",
                    "-c",
                    "sh",
                    "-c",
                    "echo $IN_NIX_SHELL",
                ]
            )
            .run()
            .ok()
        )
        assert result.stdout_s == "impure\n"


@with_files(
    {
        "t": flake | get_global_asset_pack(".git") | {".gitignore": File("flake.lock\n")},
        "nixpkgs": nixpkgs,
    }
)
def test_ignoring_lockfile_does_not_break_develop(nix: Nix, files: Path, git: Git):
    git(files / "t", "add", "config.nix", "shell-hello.nix", "flake.nix", ".gitignore")

    nix.nix(["develop", ".#hello"], cwd=files / "t").with_stdin(b"true").run().ok()



from collections.abc import Callable, Generator
from typing import Literal, Any

import pytest
from testlib.utils import is_value_of_type


def test_list_type_valid():
    assert is_value_of_type([1, 2, 3], list[int])


def test_list_type_valid_multi_type():
    assert is_value_of_type([1, "a", 2], list[int | str])
    assert is_value_of_type([1, "a", 2], set[str | int] | list[int | str])


def test_list_type_invalid_single_fail():
    assert not is_value_of_type([1, 2, "a"], list[int])


def test_list_type_invalid_all_wrong():
    assert not is_value_of_type([1, 2, 3], list[str])


def test_list_type_nested():
    assert is_value_of_type([[1], [2, 3], [4]], list[list[int]])


def test_list_type_nested_single_invalid():
    assert not is_value_of_type([[1], [2, 3], ["a"]], list[list[int]])


def test_weird_type_valid():
    assert is_value_of_type(42, Literal[42])


def test_type_type():
    class Foo: ...

    assert is_value_of_type(Foo, type[Foo])


def test_type_doesnt_allow_generator():
    def foo():
        yield 1

    with pytest.raises(ValueError, match=r"Unsupported expected_type.+"):
        is_value_of_type(foo, Generator[int, None, None])


def test_type_doesnt_allow_callable():
    def foo(a):  # noqa: ANN001, ANN202 # dummy stuff within testing, nothing external
        return "a" + str(a)

    with pytest.raises(ValueError, match=r"Unsupported expected_type.+"):
        is_value_of_type(foo, Callable[[int], str])


def test_type_allows_none():
    assert is_value_of_type(None, type[None])


def test_doesnt_allow_generic():
    class X[T]: ...

    with pytest.raises(ValueError, match=r"Unsupported expected_type.+"):
        is_value_of_type(0, X[int])


def test_type_dict_checks_valid():
    d = {"a": 1, "b": 2}
    assert is_value_of_type(d, dict[str, int])


def test_type_dict_key_invalid():
    d = {"a": 1, 1: 2}
    assert not is_value_of_type(d, dict[str, int])


def test_type_dict_value_invalid():
    d = {"a": 1, "b": "2"}
    assert not is_value_of_type(d, dict[str, int])


def test_type_allows_any():
    assert is_value_of_type(1, Any)
    assert is_value_of_type("a", Any)
    assert is_value_of_type(lambda x: x, Any)



import dataclasses
import logging
import platform
import shutil
from pathlib import Path

from testlib.environ import environ

import pytest

SLASHES_IN_STORE_PATH_UNTIL_PACKAGE = "/nix/store/hash-program_name/".count("/")


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _ManagedPath:
    """
    Wrapper class to handle building the `PATH` environment variable
    """

    build_shell: dataclasses.InitVar[str | None]
    build_env: dataclasses.InitVar[str | None] = None
    """statically linked shell to use within builds which provides coreutils functionality"""
    _path: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self, build_shell: str | None, build_env: str | None):
        if build_shell:
            self.prepend(build_shell)
        if build_env:
            for part in build_env.split(":"):
                self.append(part)

    def to_path(self) -> str:
        """
        :return: string to be put into the `PATH` environment variable containing all added paths/programs
        """
        return ":".join(self._path)

    def prepend(self, exec_path: str | Path) -> "_ManagedPath":
        """
        Adds the given file or folder at the FRONT of the path variable
        :param exec_path: executable or folder containing executables to be added
        :return: self, to allow for chaining
        """
        self._path.insert(0, str(exec_path))
        return self

    def append(self, exec_path: str | Path) -> "_ManagedPath":
        """
        Adds the given file or folder at the END of the path variable
        :param exec_path: executable or folder containing executables to be added
        :return: self, to allow for chaining
        """
        self._path.append(str(exec_path))
        return self

    def insert_at(self, exec_path: str | Path, index: int) -> "_ManagedPath":
        """
        Adds the given file or folder at the GIVEN INDEX of the path variable
        :param exec_path: executable or folder containing executable to be added
        :param index: where to insert the path
        :return: self, to allow for chaining
        """
        self._path.insert(index, str(exec_path))
        return self

    def add_program(self, program_name: str, all_associated: bool = True) -> "_ManagedPath":
        """
        Adds the given program to the path by name.
        :param program_name: executable/program to add
        :param all_associated: if True, the folder containing the executable will be added instead. Otherwise, only the provided executable will be added
        :raises ValueError: if the program could not be found
        :return: self, to allow for chaining
        """
        path = shutil.which(program_name)
        if path is None:
            msg = f"Couldn't find program {program_name!r}"
            raise ValueError(msg)
        # Convert to path object for better checking and operations
        path = Path(path)
        if all_associated and not path.is_dir():
            path = path.parent
        # handle as string within the data structure
        path = str(path)
        if path not in self._path:
            self._path.append(path)
        return self

    def remove_path(self, exec_path: str | Path) -> "_ManagedPath":
        """
        Removes the given file or folder from the path
        :param exec_path: file or folder to remove
        :raises ValueError: if the file or folder could not be found
        :return: self, to allow for chaining
        """
        self._path.remove(str(exec_path))
        return self

    def remove_program(self, program_name: str) -> "_ManagedPath":
        """
        Removes the given executable/program from the path by name
        :param program_name: executable/program to remove
        :raises ValueError: if the program was not found or isn't present in path
        :return: self, to allow for chaining
        """
        path = shutil.which(program_name)
        if path is None:
            msg = f"Couldn't find program {program_name!r}"
            raise ValueError(msg)
        if path in self._path or (path := str(Path(path).parent)) in self._path:
            self.remove_path(path)
        else:
            # Mirror behavior of `remove_path`
            msg = f"path.remove({program_name}): {program_name} not in path"
            raise ValueError(msg)
        return self

    def to_sandbox_paths(self) -> list[str]:
        """
        :return: list of strings to put into the `sandbox_paths` nix setting
        """
        ret = []
        for p in self._path:
            if p.startswith("/nix/store/"):
                # adds the entire package to the sandbox,
                # to ensure that dependencies and libraries from within the package are also present
                ret.append("/".join(p.split("/")[:SLASHES_IN_STORE_PATH_UNTIL_PACKAGE]))
            else:
                ret.append(p)
        return ret

    def which(self, program_name: str) -> Path:
        path = shutil.which(program_name, path=self.to_path())
        if not path:
            raise ValueError(f"{program_name} is not in configured path")
        return Path(path)


@dataclasses.dataclass
class _Dirs:
    test_root: Path
    home: Path
    nix_log_dir: Path | None
    nix_state_dir: Path | None
    nix_conf_dir: Path | None
    nix_bin_dir: Path | None
    # this one *must not* be NIX_STORE_DIR, otherwise lix will pick it up
    # and misconfigure itself. the config system is unbelievable bullshit
    real_store_dir: Path | None
    cache_dir: Path | None
    xdg_cache_home: Path | None
    tmpdir: Path | None
    """used for nar caching"""
    nix_store_dir: Path | None = None

    def get_env_keys(self) -> set[str]:
        return {f.name.upper() for f in dataclasses.fields(self)}

    def to_env_vars(self) -> dict[str, str]:
        return {k.upper(): v for k, v in dataclasses.asdict(self).items() if v is not None}


class ManagedEnv:
    def __init__(self, tmp_path: Path):
        # Things fetched from the global env
        build_shell = environ.get("BUILD_TEST_SHELL")
        global_path = environ.get("PATH")
        build_env = environ.get("BUILD_TEST_ENV")
        # `NIX_BIN_DIR` either propagated from us or set by meson
        # Set to the codebase internal output if started standalone
        # This is where the current lix binaries are located.
        # local import to avoid cyclic dependencies
        from testlib.utils import lix_base_folder  # noqa: PLC0415

        lix_bin = Path(environ.get("NIX_BIN_DIR", lix_base_folder / "outputs/out/bin"))

        self._env = {}
        self.path = _ManagedPath(build_shell, build_env)
        self._tmp_path = tmp_path
        self.shell_dir = build_shell or "/bin"
        self.build_env = build_env

        self.dirs = _Dirs(
            test_root=self._get_dir(""),
            home=self._get_dir("test-home"),
            nix_log_dir=self._get_dir("var/log/nix"),
            nix_state_dir=self._get_dir("nix/var/nix"),
            nix_conf_dir=self._get_dir("etc/nix"),
            nix_bin_dir=lix_bin,
            real_store_dir=self._get_dir("nix/store"),
            nix_store_dir=self._get_dir("nix/store"),
            cache_dir=self._get_dir("test-binary-cache"),
            xdg_cache_home=self._get_dir("test-home/.cache"),
            tmpdir=self._get_dir("tmp"),
        )
        self.path.prepend(self.dirs.nix_bin_dir)
        self.path.prepend(self.shell_dir)
        self.init_defaults(global_path)

        self._registered_commands = []

    def _get_dir(self, sub_path: str) -> Path:
        p = self._tmp_path / sub_path
        p.mkdir(parents=True, exist_ok=True)
        return p

    def init_defaults(self, global_path: str):
        self._env = {
            # Do not use the system-wide or local config for git, but *none* instead
            "GIT_CONFIG_SYSTEM": "/dev/null",
            # Shell to use, required by lix to run sub processes / commands
            "SHELL": f"{self.shell_dir}/sh",
            # when writing things to the terminal (esp with man pages) use cat, to print the full output to stdout
            "PAGER": "cat",
            "BUILD_TEST_SHELL": self.shell_dir,
        }
        if self.build_env:
            self._env["BUILD_TEST_ENV"] = self.build_env
        if platform.system() == "Darwin":
            # Darwin / Apple behaves differently and requires _NIX_TEST_NO_SANDBOX to be set for whatever reason
            self._env |= {"_NIX_TEST_NO_SANDBOX": "1"}
            # copy global path to maintain features usually provided by busybox
            [self.path.append(p) for p in global_path.split(":") if Path(p).exists()]

    def set_env(self, name: str, value: str):
        if name in self.dirs.get_env_keys():
            msg = f"Overriding paths should be done using the `env.dirs` attribute, use `env.dirs.{name.lower()}` instead."
            raise ValueError(msg)
        if name == "PATH":
            msg = "Setting of path not supported. use `env.path` instead"
            raise ValueError(msg)
        if value is None:
            msg = "setting to `None` is not allowed. did you mean to use `env.unset_env`?"
            raise ValueError(msg)
        self._env[name] = value

    def __setitem__(self, key: str, value: str) -> None:
        return self.set_env(key, value)

    def get_env(self, name: str, default: str | Path | None = None) -> str | Path | None:
        if name in self.dirs.get_env_keys():
            return getattr(self.dirs, name.lower())
        if name == "PATH":
            msg = "getting of path not supported, use `env.path` instead"
            raise ValueError(msg)
        return self._env.get(name, default)

    def __getitem__(self, item: str) -> str | Path | None:
        itm = self.get_env(item)
        if itm is None:
            msg = f"{itm} is not set"
            raise KeyError(msg)
        return itm

    def unset_env(self, name: str) -> str | None:
        if name in self.dirs.get_env_keys():
            msg = f"Overriding paths should be done using the `env.dirs` attribute, use `env.dirs.{name.lower()}` instead."
            raise ValueError(msg)
        return self._env.pop(name, None)

    def __delitem__(self, key: str) -> str | None:
        itm = self.unset_env(key)
        if itm is None:
            msg = f"{itm} is not set"
            raise KeyError(msg)
        return itm

    def to_env(self) -> dict[str, str]:
        ret = self.dirs.to_env_vars()
        ret["PATH"] = self.path.to_path()
        for k, v in self._env.copy().items():
            if v is not None:
                ret[k] = v
            else:
                logger.warning("environment variable '%s' is none", k)
        return ret


@pytest.fixture
def env(tmp_path: Path) -> ManagedEnv:
    env_ = ManagedEnv(tmp_path)
    yield env_
    for cmd in env_._registered_commands:
        if not cmd._run:
            e = ValueError("Command has not been run")
            e.add_note(str(cmd))
            raise e



import dataclasses
import re
import shutil
import sys
from pathlib import Path

import pytest

from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.env import _ManagedPath


def test_env_inits_defaults(tmp_path: Path):
    env = ManagedEnv(tmp_path)
    assert env.get_env("GIT_CONFIG_SYSTEM") == "/dev/null"
    if sys.platform != "darwin":
        # Darwin doesn't like sandboxes so we don't have a sandbox shell here
        assert "bash" in env.get_env("SHELL")
    else:
        assert "/bin/sh" in env.get_env("SHELL")
    assert env.get_env("PAGER") == "cat"

    assert env.dirs.test_root == tmp_path
    assert env.dirs.home == tmp_path / "test-home"
    assert env.dirs.real_store_dir == tmp_path / "nix/store"


def test_env_unknown_none(env: ManagedEnv):
    assert env.get_env("SOME_UNKNOWN_STUFF") is None


def test_env_unknown_default(env: ManagedEnv):
    assert env.get_env("QUESTION_OF_LIFE", "42") == "42"


def test_env_sets(env: ManagedEnv):
    assert env.get_env("DRGN") is None
    env.set_env("DRGN", "cute")
    assert env.get_env("DRGN") == "cute"


def test_env_setitem_success(env: ManagedEnv):
    env["HELLO"] = "world"
    assert env.get_env("HELLO") == "world"


def test_env_getitem_known(env: ManagedEnv):
    env["HELLO"] = "world"
    assert env["HELLO"] == "world"


def test_env_getitem_unknown(env: ManagedEnv):
    with pytest.raises(KeyError):
        _ = env["HELLO"]


def test_env_overrides_custom(env: ManagedEnv):
    env.set_env("VALID", "you")
    assert env.get_env("VALID") == "you"
    env.set_env("VALID", "every creature")
    assert env.get_env("VALID") == "every creature"


def test_env_overrides_defaults(env: ManagedEnv):
    assert env.get_env("PAGER") == "cat"
    env.set_env("PAGER", "bat")
    assert env.get_env("PAGER") == "bat"


def test_env_unset_custom(env: ManagedEnv):
    env.set_env("FAILURE", "me")
    assert env.get_env("FAILURE") == "me"
    env.unset_env("FAILURE")
    assert env.get_env("FAILURE") is None


def test_env_unset_default(env: ManagedEnv):
    assert env.get_env("PAGER") == "cat"
    env.unset_env("PAGER")
    assert env.get_env("PAGER") is None


def test_env_set_none_errors(env: ManagedEnv):
    with pytest.raises(ValueError, match=r".+env.unset_env.+"):
        env.set_env("PAGER", None)  # type: ignore


def test_env_set_no_dirs(env: ManagedEnv):
    with pytest.raises(ValueError, match=r"Overriding paths should be done .+"):
        env.set_env("HOME", "/home/zelda")


def test_env_get_dir_works(env: ManagedEnv):
    assert env.get_env("HOME") == env.dirs.home


def test_env_dirs_created(env: ManagedEnv):
    fields = dataclasses.asdict(env.dirs).items()
    assert len(fields) > 1
    for name, field in fields:
        field.exists()


def test_env_get_path_fails(env: ManagedEnv):
    with pytest.raises(ValueError, match=r".+env\.path.+"):
        _ = env.get_env("PATH")


def test_env_set_path_fails(env: ManagedEnv):
    with pytest.raises(ValueError, match=r".+env\.path.+"):
        env.set_env("PATH", "a:b")


def test_env_to_env(tmp_path: Path):
    env = ManagedEnv(tmp_path)
    assert set(env.to_env().keys()) == {
        "GIT_CONFIG_SYSTEM",
        "SHELL",
        "PAGER",
        "HOME",
        "TEST_ROOT",
        "NIX_LOG_DIR",
        "NIX_STATE_DIR",
        "NIX_CONF_DIR",
        "NIX_BIN_DIR",
        "NIX_STORE_DIR",
        "REAL_STORE_DIR",
        "CACHE_DIR",
        "XDG_CACHE_HOME",
        "PATH",
        "BUILD_TEST_SHELL",
        "TMPDIR",
        "BUILD_TEST_ENV",
    } | ({"_NIX_TEST_NO_SANDBOX"} if sys.platform == "darwin" else set())


def test_path_inits_build_shell():
    path = _ManagedPath("/path/to/build_shell")
    assert len(path._path) == 1
    assert path._path[0] == "/path/to/build_shell"


def test_path_appends_at_end():
    path = _ManagedPath("/path/to/build_shell")
    path.append("other_path")
    assert len(path._path) == 2
    assert path._path[1] == "other_path"


def test_path_prepends_at_front():
    path = _ManagedPath("/path/to/build_shell")
    path.prepend("other/path")
    assert len(path._path) == 2
    assert path._path[0] == "other/path"


def test_remove_path_known():
    path = _ManagedPath("/path/to/build_shell")
    path.append("other/path")
    assert len(path._path) == 2
    path.remove_path("other/path")
    assert path._path == ["/path/to/build_shell"]


def test_remove_path_unknown():
    path = _ManagedPath("/path/to/build_shell")
    path.append("other/path")
    assert len(path._path) == 2
    with pytest.raises(ValueError, match=r".+ x not in list"):
        path.remove_path("/home/nest/builder")
    assert len(path._path) == 2


def test_path_add_program_resolves():
    path = _ManagedPath("/path/to/build_shell")
    pytest_path = shutil.which("pytest")
    path.add_program("pytest", False)
    assert len(path._path) == 2
    assert path._path[1] == pytest_path


def test_path_add_program_entire_dir():
    path = _ManagedPath("/path/to/build_shell")
    pytest_path = shutil.which("pytest")
    path.add_program("pytest", True)
    assert len(path._path) == 2
    actual = path._path[1]
    assert actual != pytest_path
    assert actual == pytest_path.rpartition("/")[0]


def test_path_remove_program_specific():
    path = _ManagedPath("/path/to/build_shell")
    path.add_program("pytest", False)
    assert len(path._path) == 2
    path.remove_program("pytest")
    assert len(path._path) == 1


def test_path_remove_program_associated():
    path = _ManagedPath("/path/to/build_shell")
    path.add_program("pytest", True)
    assert len(path._path) == 2
    path.remove_program("pytest")
    assert len(path._path) == 1


def test_path_to_path():
    path = _ManagedPath("/path/to/build_shell")
    path.prepend("pictures/latest/kate")
    path.prepend("out/bin/lix")
    path.append("fops/with/noms")
    assert path.to_path() == "out/bin/lix:pictures/latest/kate:/path/to/build_shell:fops/with/noms"


def test_path_sandbox_entire_store_package():
    path = _ManagedPath("/path/to/build_shell")
    path.add_program(shutil.which("pytest"))
    sb_paths = path.to_sandbox_paths()
    assert len(sb_paths) == 2
    assert sb_paths[0] == "/path/to/build_shell"
    assert re.fullmatch(r"^/nix/store/\w{32}-python3-3\.\d{1,2}\.\d{1,2}-[^/]+$", sb_paths[1])


def test_path_nontrivial_build_env():
    path = _ManagedPath("/path/to/build_shell", "/path/to/busybox/bin:/path/to/acl/bin")
    sb_paths = path.to_sandbox_paths()
    assert len(sb_paths) == 3
    assert sb_paths[0] == "/path/to/build_shell"


def test_path_which():
    path = _ManagedPath("/path/to/build_shell")
    with pytest.raises(ValueError, match="not in configured path"):
        path.which("pytest")
    wanted = shutil.which("pytest")
    path.add_program(wanted)
    assert str(path.which("pytest")) == wanted



from pathlib import Path
import pytest

from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import File, with_files


@with_files({"example.txt": File("example")})
class TestLocalStore:
    @pytest.fixture(autouse=True)
    def setup(self, nix: Nix, files: Path):  # noqa: ARG002
        nix.settings.add_xp_feature("nix-command")

        store = nix.env.dirs.home / "x"
        store.mkdir()
        nix.env.dirs.nix_store_dir = store
        nix.env.dirs.real_store_dir = store

        self.path = (
            nix.nix_store(["--store", "./x", "--add", "example.txt"]).run().ok().stdout_plain
        )

    def test_path_info_relative(self, nix: Nix):
        store = "./x"
        path = nix.nix(["path-info", "--store", store, self.path]).run().ok().stdout_plain
        assert path == self.path

    def test_path_info_absolute(self, nix: Nix):
        store = f"{nix.env.dirs.home}/x"
        path = nix.nix(["path-info", "--store", store, self.path]).run().ok().stdout_plain
        assert path == self.path

    def test_path_info_uri(self, nix: Nix):
        store = f"local?root={nix.env.dirs.home}/x"
        path = nix.nix(["path-info", "--store", store, self.path]).run().ok().stdout_plain
        assert path == self.path

    def test_local_is_trusted(self, nix: Nix):
        info = nix.nix(["--store", "./x", "store", "ping", "--json"]).run().json()
        assert info["trusted"]

    def test_doctor_shows_trust(self, nix: Nix):
        result = nix.nix(["--store", "./x", "doctor"]).run().ok()
        assert "You are trusted by" in result.stderr_plain



import re
import json
import pytest
from textwrap import dedent
from pathlib import Path

from testlib.fixtures.file_helper import File, with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset, get_global_asset_pack
from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.command import Command
from testlib.environ import environ

system = environ.get("system")

flake = {
    "flake.nix": File("""
        {
          description = "Bla bla";

          outputs = { self }: with import ./config.nix; rec {
            packages.${system}.default = mkDerivation {
              name = "profile-test-${builtins.readFile ./version}";
              outputs = [ "out" "man" "dev" ];
              builder = builtins.toFile "builder.sh"
                ''
                  mkdir -p $out/bin
                  cat > $out/bin/hello <<EOF
                  #! ${shell}
                  echo Hello ${builtins.readFile ./who}
                  EOF
                  chmod +x $out/bin/hello
                  echo DONE
                  mkdir -p $man/share/man
                  mkdir -p $dev/include
                '';
              __contentAddressed = import ./ca.nix;
              outputHashMode = "recursive";
              outputHashAlgo = "sha256";
              meta.outputsToInstall = [ "out" "man" ];
            };
          };
        }
    """),
    "who": File("World"),
    "version": File("1.0"),
    "ca.nix": File("false"),
    "config.nix": get_global_asset("config.nix"),
}


@pytest.fixture(autouse=True)
def configure_system(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")


def run_hello(env: ManagedEnv) -> str:
    return Command([str(env.dirs.home / ".nix-profile/bin/hello")], _env=env).run().ok().stdout_s


def build(nix: Nix, thing: str) -> str:
    return nix.nix(["build", "--no-link", "--print-out-paths", thing]).run().ok().stdout_s.strip()


class TestUpgradeProfile:
    @with_files({"flake1": flake} | get_global_asset_pack("user-envs"))
    def test_nix_env_upgrade(self, nix: Nix, env: ManagedEnv):
        nix.nix_env(["-f", "./user-envs.nix", "-i", "foo-1.0"]).run().ok()
        result = nix.nix(["profile", "list"]).run().ok().stdout_s
        assert re.match(r"Name:.*foo\nStore paths:.*foo-1.0", result)

        nix.nix(["profile", "install", "./flake1", "-L"]).run().ok()
        result = nix.nix(["profile", "list"]).run().ok().stdout_s
        assert re.search(r"Name:.*flake1\n.*?\n.*?\nLocked flake URL:.*narHash", result)

        assert run_hello(env) == "Hello World\n"
        assert (env.dirs.home / ".nix-profile/share/man").exists()
        assert not (env.dirs.home / ".nix-profile/include").exists()

        result = nix.nix(["profile", "history"]).run().ok().stdout_s
        assert f"packages.{system}.default: ∅ -> 1.0" in result

        result = nix.nix(["profile", "diff-closures"]).run().ok().stdout_s
        assert "env-manifest.nix: ε → ∅" in result

    @with_files({"flake1": flake})
    def test_upgrade_from_v2(self, nix: Nix, env: ManagedEnv):
        import_profile = env.dirs.home / "import-profile"
        import_profile.mkdir()

        path = build(nix, "./flake1^out")
        (import_profile / "manifest.json").write_text(
            json.dumps(
                {
                    "version": 2,
                    "elements": [
                        {
                            "active": True,
                            "attrPath": "legacyPackages.x86_64-linux.hello",
                            "originalUrl": "flake:nixpkgs",
                            "outputs": None,
                            "priority": 5,
                            "storePaths": [path],
                            "url": "github:NixOS/nixpkgs/aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                        }
                    ],
                }
            )
        )

        added_path = nix.nix(["store", "add-path", str(import_profile)]).run().ok().stdout_s.strip()
        nix.nix(
            ["build", "--profile", str(env.dirs.home / ".nix-profile"), str(added_path)]
        ).run().ok()

        result = nix.nix(["profile", "list"]).run().ok().stdout_s
        assert re.match("Name:.*hello\n.*\nStore paths:.*" + path, result, re.S)
        assert (
            "removed 1 packages, kept 0 packages"
            in nix.nix(["profile", "remove", "hello"]).run().ok().stderr_s
        )


class ProfileTestBase:
    @pytest.fixture(autouse=True)
    def setup(self, nix: Nix, configure_system, files):  # noqa: ANN001, ARG002
        nix.nix(["profile", "install", "./flake1", "-L"]).run().ok()


class TestBasicOperations(ProfileTestBase):
    @with_files({"flake1": flake})
    def test_xdg_base_dir_support(self, nix: Nix, env: ManagedEnv):
        nix.settings.use_xdg_base_directories = True

        result = nix.nix(["profile", "remove", "flake1", "-L"]).run().ok()
        assert "removed 1 packages" in result.stderr_s

        nix.nix(["profile", "install", "./flake1", "-L"]).run().ok()
        assert run_hello(env) == "Hello World\n"

    @with_files({"flake1": flake})
    def test_wipe_history(self, nix: Nix):
        nix.nix(["profile", "wipe-history"]).run().ok()
        history = nix.nix(["profile", "history"]).run().ok().stdout_s
        assert len(re.findall(r"^Version ", history, re.M)) == 1

    @with_files({"flake1": flake} | get_global_asset_pack("simple-drv"))
    def test_install_non_flake(self, nix: Nix, env: ManagedEnv):
        nix.nix(["profile", "install", "--file", "./simple.nix", ""]).run().ok()
        assert (env.dirs.home / ".nix-profile/hello").read_text() == "Hello World!\n"
        assert "removed 1 packages" in nix.nix(["profile", "remove", "simple"]).run().ok().stderr_s

        assert not (env.dirs.home / ".nix-profile/hello").exists()
        output = nix.nix_build(["--no-out-link", "./simple.nix"]).run().ok().stdout_s.strip()
        nix.nix(["profile", "install", output]).run().ok()
        assert (env.dirs.home / ".nix-profile/hello").read_text() == "Hello World!\n"

    @with_files(
        {
            "flake1": flake,
            "simple1": get_global_asset_pack("simple-drv"),
            "simple2": get_global_asset_pack("simple-drv"),
        }
    )
    def test_install_different_sources(self, nix: Nix):
        nix.nix(["profile", "install", "--file", "./simple1/simple.nix", ""]).run().ok()
        nix.nix(["profile", "install", "--file", "./simple2/simple.nix", ""]).run().ok()

        result = nix.nix(["profile", "list"]).run().ok().stdout_s
        assert re.search(r"Name:.*simple", result)
        assert re.search(r"Name:.*simple-1", result)

        assert "removed 1 packages" in nix.nix(["profile", "remove", "simple"]).run().ok().stderr_s
        assert (
            "removed 1 packages" in nix.nix(["profile", "remove", "simple-1"]).run().ok().stderr_s
        )

    @with_files({"flake1": flake, "flake2": flake | {"who": File("World2")}})
    def test_priority(self, nix: Nix, env: ManagedEnv):
        hello1 = build(nix, "./flake1#default.out")
        hello2 = build(nix, "./flake2#default.out")

        result = nix.nix(["--offline", "profile", "install", "./flake2"]).run().expect(1)
        result = [
            line
            for line in result.stderr_s.splitlines()
            if not line.startswith("warning: ")
            and not line.startswith("error (ignored): ")
            and not re.match(r"^fetching .+ input", line)
            and line
        ]
        assert result == [
            "error: An existing package already provides the following file:",
            f"         {hello1}/bin/hello",
            "       This is the conflicting file from the new package:",
            f"         {hello2}/bin/hello",
            "       To remove the existing package:",
            f"         nix profile remove path:{env.dirs.home}/flake1#packages.{system}.default",
            "       The new package can also be installed next to the existing one by assigning a different priority.",
            "       The conflicting packages have a priority of 5.",
            "       To prioritise the new package:",
            f"         nix profile install path:{env.dirs.home}/flake2#packages.{system}.default --priority 4",
            "       To prioritise the existing package:",
            f"         nix profile install path:{env.dirs.home}/flake2#packages.{system}.default --priority 6",
        ]

        assert run_hello(env) == "Hello World\n"
        nix.nix(["profile", "install", "./flake2", "--priority", "100"]).run().ok()
        assert run_hello(env) == "Hello World\n"
        nix.nix(["profile", "install", "./flake2", "--priority", "0"]).run().ok()
        assert run_hello(env) == "Hello World2\n"
        nix.nix(["profile", "install", "./flake1", "--priority", "100"]).run().ok()
        assert run_hello(env) == "Hello World2\n"


@with_files({"flake1": flake})
class TestProfileContentUpgrade(ProfileTestBase):
    @pytest.fixture(autouse=True)
    def upgrade(self, nix: Nix, env: ManagedEnv, setup):  # noqa: ANN001, ARG002
        (env.dirs.home / "flake1/who").write_text("NixOS")
        (env.dirs.home / "flake1/version").write_text("2.0")

        nix.nix(["profile", "upgrade", "flake1"]).run().ok()

    def test_upgrade_succeeded(self, nix: Nix, env: ManagedEnv):
        assert run_hello(env) == "Hello NixOS\n"

        result = nix.nix(["profile", "history"]).run().ok().stdout_s
        assert f"packages.{system}.default: 1.0, 1.0-man -> 2.0, 2.0-man" in result

    def test_diff_closures_works(self, nix: Nix):
        assert nix.nix(["profile", "diff-closures"]).run().stdout_s == dedent("""\
            Version 1 -> 2:
              profile-test: 1.0 → 2.0
        """)

    def test_rollback_works(self, nix: Nix, env: ManagedEnv):
        nix.nix(["profile", "rollback"]).run().ok()
        assert run_hello(env) == "Hello World\n"


@with_files({"flake1": flake} | get_global_asset_pack("user-envs"))
def test_uninstall(nix: Nix, env: ManagedEnv):
    nix.nix_env(["-f", "./user-envs.nix", "-i", "foo-1.0"]).run().ok()
    nix.nix(["profile", "install", "./flake1", "-L"]).run().ok()

    assert (env.dirs.home / ".nix-profile/bin/foo").exists()
    assert "removed 1 packages" in nix.nix(["profile", "remove", "foo"]).run().ok().stderr_s
    assert not (env.dirs.home / ".nix-profile/bin/foo").exists()
    assert "foo: 1.0 -> ∅" in nix.nix(["profile", "history"]).run().ok().stdout_s
    assert "Version 1 -> 2" in nix.nix(["profile", "diff-closures"]).run().ok().stdout_s


@with_files({"flake1": flake})
def test_output_override(nix: Nix, env: ManagedEnv):
    profile = env.dirs.home / ".nix-profile"

    nix.nix(["profile", "install", "./flake1^*"]).run().ok()
    assert run_hello(env) == "Hello World\n"
    assert (profile / "share/man").exists()
    assert (profile / "include").exists()

    assert (env.dirs.home / "flake1/who").write_text("Lix")
    nix.nix(["profile", "upgrade", "flake1"]).run().ok()
    assert run_hello(env) == "Hello Lix\n"
    assert (profile / "share/man").exists()
    assert (profile / "include").exists()

    assert "removed 1 packages" in nix.nix(["profile", "remove", "flake1"]).run().ok().stderr_s
    nix.nix(["profile", "install", "./flake1^man"]).run().ok()
    assert not (profile / "bin/hello").exists()
    assert (profile / "share/man").exists()
    assert not (profile / "include").exists()


@with_files({"flake1": flake, "flake2": flake})
def test_conflict_resolution_cppnix_8284(nix: Nix):
    path = build(nix, "./flake1^out")
    nix.nix(["profile", "install", path]).run().ok()

    expr = f'(builtins.getFlake "./flake2").packages.{system}.default'
    nix.nix(["profile", "install", "--impure", "--expr", expr]).run().expect(1)


class TestKeepEnvDerivations:
    @with_files({"flake1": flake})
    def test_keep_derivations_enabled(self, nix: Nix, env: ManagedEnv):
        nix.settings["keep-env-derivations"] = "true"
        nix.settings["keep-derivations"] = "false"  # NOTE: This defaults to `true`.

        nix.nix(["profile", "install", "./flake1", "-L"]).run().ok()
        assert run_hello(env) == "Hello World\n"
        res = nix.nix(["eval", "--raw", "./flake1#default.drvPath"]).run().ok()
        drv_path = Path(res.stdout_s.strip())

        assert drv_path.exists()

        # The derivation should NOT be collected because `keep-env-derivations` is true
        nix.nix(["store", "gc"]).run().ok()

        assert drv_path.exists()

    @with_files({"flake1": flake})
    def test_keep_derivations_disabled(self, nix: Nix, env: ManagedEnv):
        nix.settings["keep-derivations"] = "false"  # NOTE: This defaults to `true`

        nix.nix(["profile", "install", "./flake1", "-L"]).run().ok()
        assert run_hello(env) == "Hello World\n"

        res = nix.nix(["eval", "--raw", "./flake1#default.drvPath"]).run().ok()
        drv_path = Path(res.stdout_s.strip())

        assert drv_path.exists()

        # The derivation should be collected because `keep-env-derivations` is false AND `keep-derivations` is false.
        res = nix.nix(["store", "gc"]).run().ok()

        assert not drv_path.exists()



import builtins
import types
import typing
from pathlib import Path
from textwrap import dedent
from types import UnionType
from typing import Any, Literal, get_args, get_origin

from testlib.environ import environ
from testlib.fixtures.file_helper import (
    CopyFile,
    CopyTree,
    FileDeclaration,
    merge_file_declaration,
    Fileish,
    CopyTemplate,
    File,
)

# Things have to be resolved from top to bottom, because otherwise the tests behave flakey
# due to the internal file structure

testlib_folder = Path(__file__).parent.absolute()

global_assets_folder = testlib_folder / "global_assets"

functional2_base_folder = testlib_folder.parent
"""
Resolves to `tests/functional2` folder
"""


test_base_folder = functional2_base_folder.parent
"""
Resolves the base path to the `tests` folder of the project
This ensures that the relative path all start with functional2.
"""

lix_base_folder = test_base_folder.parent
"""
Resolves to the base path of the lix git repository
"""


def get_functional2_files(additional_files: FileDeclaration | None = None) -> FileDeclaration:
    """
    Returns the very basic Configuration of functional2 (i.e. its init and pyproject toml file)
    Note that this does not include the conftest.py nor anything from testlib
    as those might want to be configured manually
    :param additional_files: any additional files, one wants to have
    :return: file declaration to be used for the files fixtures to set up functional2 within a temp folder
    """
    if additional_files is None:
        additional_files = {}
    return merge_file_declaration(
        {"functional2": {"pyproject.toml": CopyFile(functional2_base_folder / "pyproject.toml")}},
        additional_files,
    )


def get_functional2_files_with_testlib(
    additional_files: FileDeclaration | None = None, no_tests: bool = True
) -> FileDeclaration:
    if additional_files is None:
        additional_files = {}
    if no_tests:

        def ignore(_: Any, names: list[str]) -> list[str]:
            return [name for name in names if name.startswith("test_")]
    else:
        ignore = None
    return get_functional2_files(
        merge_file_declaration(
            {
                "functional2": {
                    "conftest.py": CopyFile(functional2_base_folder / "conftest.py"),
                    "testlib": CopyTree(functional2_base_folder / "testlib", ignore=ignore),
                }
            },
            additional_files,
        )
    )


def get_functional2_lang_files(additional_files: FileDeclaration | None = None) -> FileDeclaration:
    if additional_files is None:
        additional_files = {}
    return get_functional2_files_with_testlib(
        merge_file_declaration(
            {
                "functional2": {
                    "lang": {
                        "__init__.py": CopyFile(functional2_base_folder / "lang/__init__.py"),
                        "lang_util.py": CopyFile(functional2_base_folder / "lang/lang_util.py"),
                        "test_lang.py": CopyFile(functional2_base_folder / "lang/test_lang.py"),
                        "lib.nix": CopyFile(functional2_base_folder / "lang/lib.nix"),
                    }
                }
            },
            additional_files,
        )
    )


def is_value_of_type(value: Any, expected_type: type[Any] | UnionType) -> bool:
    """
    checks if the given value conforms to the type.
    This function is useful, to check if all values conform to the inner iterable type
    :param value: value to check
    :param expected_type: what type each item should be; must be a non-generic type or one of {dict, list, set, UnionType, Literal}
    :return: True if it conforms, otherwise False
    """
    supported_origin_types = {dict, list, set, type, UnionType, Literal}
    origin = get_origin(expected_type)
    if expected_type is Any:
        return True
    match origin:
        case None:
            return isinstance(value, expected_type)
        case types.UnionType:
            return any(is_value_of_type(value, t) for t in get_args(expected_type))
        case typing.Literal:
            return value in get_args(expected_type)
        case builtins.type:
            return value is get_args(expected_type)[0]
        case builtins.list | builtins.set | builtins.dict:
            matches_origin = isinstance(value, origin)
            if not matches_origin:
                return False
            matches = all(is_value_of_type(v, get_args(expected_type)[0]) for v in value)
            if origin is dict:
                # also check the value side of dicts
                matches = matches and all(
                    is_value_of_type(v, get_args(expected_type)[1]) for v in value.values()
                )
            return matches
        case _:
            msg = f"Unsupported expected_type. Must be a non-generic or one of {supported_origin_types!r}"
            raise ValueError(msg)


def get_global_asset(name: str) -> Fileish:
    if name == "config.nix":
        return CopyTemplate(
            functional2_base_folder / "testlib" / "global_assets" / "config.nix.template",
            {
                "system": environ.get("system"),
                # Either just the build shell or entire global path if we are darwin
                "path": environ.get("BUILD_TEST_ENV") or environ.get("PATH"),
                "shell": Path(environ.get("BUILD_TEST_SHELL") or "/bin") / "bash",
            },
        )
    return CopyFile(functional2_base_folder / "testlib" / "global_assets" / name)


def get_global_asset_pack(name: str) -> FileDeclaration:
    def folder_to_assets(folder: str) -> FileDeclaration:
        return {
            f.name: get_global_asset(f"{folder}/{f.name}")
            for f in (global_assets_folder / folder).iterdir()
        }

    match name:
        case ".git":
            return {
                ".git": {
                    "config": File(
                        dedent("""
                    [core]
                        repositoryformatversion = 0
                        filemode = true
                        bare = false
                        logallrefupdates = true

                    [user]
                        email = "tests+functional2@lix.systems"
                        name = "functional2"
                """)
                    ),
                    "description": File(
                        "Unnamed repository; edit this file 'description' to name the repository."
                    ),
                    "HEAD": File("ref: refs/heads/main"),
                    "info": {"exclude": File("")},
                    "objects": {"info": {}, "pack": {}},
                    "refs": {"heads": {}, "tags": {}},
                }
            }
        # default case, if its just config.nix + folder content
        case _ if (global_assets_folder / name).exists():
            return {"config.nix": get_global_asset("config.nix")} | folder_to_assets(name)
        case _:
            msg = f"invalid pack name {name!r}"
            raise ValueError(msg)



import pytest

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset


_mo_files = {
    "multiple-outputs.nix": get_global_asset("multiple-outputs.nix"),
    "config.nix": get_global_asset("config.nix"),
}
_build_args = ["build", "-f", "multiple-outputs.nix", "--no-link", "--json"]


@with_files(_mo_files)
def test_build_outputs(nix: Nix):
    res = nix.nix([*_build_args, "a", "b"], flake=True).run().ok()
    a, b = res.json()
    assert a["drvPath"].endswith("multiple-outputs-a.drv")
    assert len(a["outputs"]) == 2
    assert a["outputs"]["first"].endswith("multiple-outputs-a-first")
    assert a["outputs"]["second"].endswith("multiple-outputs-a-second")

    assert b["drvPath"].endswith("multiple-outputs-b.drv")
    assert len(b["outputs"]) == 1
    assert b["outputs"]["out"].endswith("multiple-outputs-b")


@with_files(_mo_files)
def test_build_circumflex_0(nix: Nix):
    res = nix.nix([*_build_args, "a^first"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-a.drv")
    assert a["outputs"].keys() == {"first"}


@with_files(_mo_files)
def test_build_circumflex_1(nix: Nix):
    res = nix.nix([*_build_args, "a^second,first"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-a.drv")
    assert a["outputs"].keys() == {"first", "second"}


@with_files(_mo_files)
def test_build_circumflex_2(nix: Nix):
    res = nix.nix([*_build_args, "a^*"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-a.drv")
    assert a["outputs"].keys() == {"first", "second"}


@with_files(_mo_files)
def test_build_outputs_to_install(nix: Nix):
    res = nix.nix([*_build_args, "e"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-e.drv")
    assert a["outputs"].keys() == {"a_a", "b"}


@with_files(_mo_files)
def test_build_outputs_to_install_overridden_0(nix: Nix):
    res = nix.nix([*_build_args, "e^a_a"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-e.drv")
    assert a["outputs"].keys() == {"a_a"}


@with_files(_mo_files)
def test_build_outputs_to_install_overridden_1(nix: Nix):
    res = nix.nix([*_build_args, "e^*"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-e.drv")
    assert a["outputs"].keys() == {"a_a", "b", "c"}


@with_files(_mo_files)
def test_build_non_drv(nix: Nix):
    res = nix.nix([*_build_args, "e.a_a.outPath"], flake=True).run().ok()
    a = res.json()[0]
    assert a["drvPath"].endswith("multiple-outputs-e.drv")
    assert a["outputs"].keys() == {"a_a"}


@with_files(_mo_files)
def test_build_illegal_type(nix: Nix):
    res = nix.nix([*_build_args, "e.a_a.drvPath"], flake=True).run().expect(1)
    assert (
        "has a context which refers to a complete source and binary closure. This is not supported at this time"
        in res.stderr_s
    )


@with_files(_mo_files)
def test_build_no_context(nix: Nix):
    res = nix.nix(["build", "--expr", '""', "--no-link"], flake=True).run().expect(1)
    assert (
        "error: string '' has 0 entries in its context. It should only have exactly one entry"
        in res.stderr_s
    )


@with_files(_mo_files)
def test_build_too_much_context(nix: Nix):
    res = (
        nix.nix(
            [
                "build",
                "--impure",
                "--expr",
                'with (import ./multiple-outputs.nix).e.a_a; "${drvPath}${outPath}"',
            ],
            flake=True,
        )
        .run()
        .expect(1)
    )
    assert "has 2 entries in its context. It should only have exactly one entry" in res.stderr_s


@with_files(_mo_files)
def test_build_unsafe_discard(nix: Nix):
    res = (
        nix.nix(
            [
                "build",
                "--impure",
                "--json",
                "--expr",
                "builtins.unsafeDiscardOutputDependency (import ./multiple-outputs.nix).e.a_a.drvPath",
                "--no-link",
            ],
            flake=True,
        )
        .run()
        .ok()
    )
    assert res.json()[0].endswith("multiple-outputs-e.drv")


@pytest.fixture
def drv(nix: Nix) -> str:
    nix.settings.add_xp_feature("nix-command")
    return (
        nix.nix(["eval", "-f", "multiple-outputs.nix", "--raw", "a.drvPath"])
        .run()
        .ok()
        .stdout_plain
    )


@with_files(_mo_files)
def test_build_not_an_output(nix: Nix, drv: str):
    res = nix.nix(["build", f"{drv}^not-an-output", "--no-link", "--json"]).run().expect(1)
    assert "does not have wanted outputs 'not-an-output'" in res.stderr_plain


@with_files(_mo_files)
def test_build_empty_output_list(nix: Nix, drv: str):
    res = nix.nix(["build", f"{drv}^", "--no-link", "--json"]).run().expect(1)
    assert "invalid extended outputs specifier" in res.stderr_plain


@with_files(_mo_files)
def test_build_star_entire_string(nix: Nix, drv: str):
    res = nix.nix(["build", f"{drv}^*nope", "--no-link", "--json"]).run().expect(1)
    assert "invalid extended outputs specifier" in res.stderr_plain


@with_files(_mo_files)
def test_build_drv_first(nix: Nix, drv: str):
    res = nix.nix(["build", f"{drv}^first", "--no-link", "--json"]).run().ok()
    out = res.json()[0]
    assert out["drvPath"].endswith("multiple-outputs-a.drv")
    assert out["outputs"].keys() == {"first"}
    assert out["outputs"]["first"].endswith("multiple-outputs-a-first")


@with_files(_mo_files)
def test_build_drv_first_second(nix: Nix, drv: str):
    res = nix.nix(["build", f"{drv}^first,second", "--no-link", "--json"]).run().ok()
    out = res.json()[0]
    assert out["drvPath"].endswith("multiple-outputs-a.drv")
    assert out["outputs"].keys() == {"first", "second"}
    assert out["outputs"]["first"].endswith("multiple-outputs-a-first")
    assert out["outputs"]["second"].endswith("multiple-outputs-a-second")


@with_files(_mo_files)
def test_build_drv_star(nix: Nix, drv: str):
    res = nix.nix(["build", f"{drv}^*", "--no-link", "--json"]).run().ok()
    out = res.json()[0]
    assert out["drvPath"].endswith("multiple-outputs-a.drv")
    assert out["outputs"].keys() == {"first", "second"}
    assert out["outputs"]["first"].endswith("multiple-outputs-a-first")
    assert out["outputs"]["second"].endswith("multiple-outputs-a-second")


@with_files(_mo_files)
def test_build_impure(nix: Nix):
    res = (
        nix.nix(
            ["build", "--impure", "-f", "multiple-outputs.nix", "--json", "e", "--no-link"],
            flake=True,
        )
        .run()
        .ok()
    )
    out = res.json()[0]
    assert out["drvPath"].endswith("multiple-outputs-e.drv")
    assert out["outputs"].keys() == {"a_a", "b"}


def test_build_stdin_empty(nix: Nix):
    res = (
        nix.nix(["build", "--no-link", "--stdin", "--json"], flake=True).with_stdin(b"").run().ok()
    )
    assert res.stdout_plain == "[]"


@with_files(_mo_files)
def test_build_stdin_stuff(nix: Nix, drv: str):
    cmd = nix.nix(["build", "--no-link", "--stdin", "--json"])
    res = cmd.with_stdin(f"{drv}^*\n".encode()).run().ok()
    out = res.json()[0]
    assert out["drvPath"].endswith("multiple-outputs-a.drv")



#!/usr/bin/env python3
"""
Preprocesses mdbook markdown, primarily for include directives.

The include directive format is as follows:
    {{#include foo/bar/baz.md}}

The content of includes will be indented as much as the directive itself.

Including a generated file (from building Lix; generally for the 'new' CLI):
    {{#include @generated@/foo/bar/baz.md}}
"""

from pathlib import Path
import json
import os, os.path
import sys
import textwrap

name = 'substitute.py'

def log(*args, **kwargs):
    kwargs['file'] = sys.stderr
    return print(f'{name}:', *args, **kwargs)

def remove_prefix_if_present(s: str, prefix: str) -> str | None:
    if s.startswith(prefix):
        return s.removeprefix(prefix)
    else:
        return None

def do_include(content: str, relative_md_path: Path, source_root: Path, search_path: Path):
    assert not relative_md_path.is_absolute(), f'{relative_md_path=} from mdbook should be relative'

    md_path_abs = source_root / relative_md_path
    var_abs = md_path_abs.parent
    assert var_abs.is_dir(), f'supposed directory {var_abs} is not a directory (cwd={os.getcwd()})'

    lines = []
    for l in content.splitlines(keepends=True):
        if remain := remove_prefix_if_present(l.strip(), "{{#include"):
            requested = remain[1:-2]
            # We indent bodies of indent directives by the indent of the
            # directive itself.
            num_leading_indent = len(l) - len(l.lstrip())

            if subpath := remove_prefix_if_present(requested, "@generated@/"):
                included = search_path / Path(subpath)
                requested = included.relative_to(search_path)
            else:
                included = source_root / relative_md_path.parent / requested
                requested = included.resolve().relative_to(source_root)
            assert included.exists(), f"{requested} not found at {included}"

            lines.append(
                textwrap.indent(
                    do_include(
                        included.read_text(),
                        requested,
                        source_root,
                        search_path,
                    ),
                    " " * num_leading_indent
                )
                + "\n"
            )
        else:
            lines.append(l)
    return "".join(lines)

def recursive_replace(data, book_root, search_path):
    match data:
        # XXX FUTURE: drop sections once mdBook is at 0.5.0 or above in nixpkgs
        case {'sections': sections}:
            return data | dict(
                sections = [recursive_replace(section, book_root, search_path) for section in sections],
            )
        case {'items': items}:
            return data | dict(
                items = [recursive_replace(item, book_root, search_path) for item in items],
            )
        case {'Chapter': chapter}:
            path_to_chapter = Path(chapter['path'])
            chapter_content = chapter['content']

            return data | dict(
                Chapter = chapter | dict(
                    # first process includes. this must happen before docroot processing since
                    # mdbook does not see these included files, only the final agglomeration.
                    content = do_include(
                        chapter_content,
                        path_to_chapter,
                        book_root,
                        search_path
                    ).replace(
                        '@docroot@',
                        ("../" * len(path_to_chapter.parent.parts) or "./")[:-1]
                    ).replace(
                        # this replacement is to avoid corrupting the
                        # hacking.md manual section on docroot
                        '@\\docroot\\@',
                        '@docroot@',
                    ),
                    sub_items = [
                        recursive_replace(sub_item, book_root, search_path)
                        for sub_item in chapter['sub_items']
                    ],
                ),
            )

        case rest:
            assert False, f'should have been called on a dict, not {type(rest)=}\n\t{rest=}'

def main():

    if len(sys.argv) > 1 and sys.argv[1] == 'supports':
        return 0

    # mdbook communicates with us over stdin and stdout.
    # It splorks us a JSON array, the first element describing the context,
    # the second element describing the book itself,
    # and then expects us to send it the modified book JSON over stdout.

    context, book = json.load(sys.stdin)

    # book_root is the directory where book contents leave (ie, src/)
    book_root = Path(context['root']) / context['config']['book'].get('src', 'src')

    # includes pointing into @generated@ will look here
    search_path = Path(os.environ['MANUAL_SUBSTITUTE_SEARCH'])

    # Find @var@ in all parts of our recursive book structure.
    replaced_content = recursive_replace(book, book_root, search_path)

    replaced_content_str = json.dumps(replaced_content)

    # Give mdbook our changes.
    print(replaced_content_str)

try:
    sys.exit(main())
except AssertionError as e:
    print(f'{name}: INTERNAL ERROR in mdbook preprocessor', file=sys.stderr)
    print(f'this is a bug in {name}', file=sys.stderr)
    raise
